/*

    File: analyse.c

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#include <stdio.h>
#ifndef  DJGPP
#include <sys/types.h>
#endif
#include "types.h"
#include "common.h"
#include "fnctdsk.h" /* get_LBA_part */
#include "hdaccess.h"
#include "analyse.h"
#include "bfs.h"
#include "swap.h"
#include "bsd.h"
#include "fat.h"
#include "ext2.h"
#include "netware.h"
#include "interface.h"
#include "rfs.h"

int search_type_0(t_param_disk_cst disk_car,t_diskext partition,int debug, int dump_ind)
{
  uchar buffer[8*0x200];
  loff_t hd_offset=partition->lba;
  partition->part_size=0;
  partition->part_type=0;
  if(hd_read_LBA(disk_car,8, &buffer, hd_offset))
    return -1;
  if((LBA2sector(disk_car,partition->lba)==disk_car->CHS.sector) && (LBA2head(disk_car,partition->lba)==disk_car->CHS.head))
  {
    if(!analyse_NTFS(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind))
      return 1;
    return 0;
  }
  if(!test_lin_swap((union swap_header *)&buffer,partition)) return 1;
  if(!analyse_FAT(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind))
    return 1;
  if(!test_netware((struct disk_netware *)&buffer,partition)) return 1;
  return 0;
}

int search_type_1(t_param_disk_cst disk_car,t_diskext partition,int debug, int dump_ind)
{
  uchar buffer[8*0x200];
  loff_t hd_offset=partition->lba;
  partition->part_size=0;
  partition->part_type=0;
  if(hd_read_LBA(disk_car,8, &buffer, hd_offset))
    return -1;
  if(!test_bsd(disk_car,(struct disklabel *)&buffer[0x200],partition,debug)) return 1;
  if(!test_beos((struct disk_super_block *)&buffer[0x200],partition)) return 1;
  return 0;
}

int search_type_2(t_param_disk_cst disk_car,t_diskext partition,int debug, int dump_ind)
{
  uchar buffer[8*0x200];
  loff_t hd_offset=partition->lba;
  partition->part_size=0;
  partition->part_type=0;
  if(hd_read_LBA(disk_car,8, &buffer, hd_offset))
    return -1;
  if(!test_linux(disk_car,(struct ext2_super_block*)&buffer[0x400],partition,debug,dump_ind)) return 1;
  return 0;
}

int search_type_128(t_param_disk_cst disk_car,t_diskext partition,int debug, int dump_ind)
{
  uchar buffer[8*0x200];
  loff_t hd_offset=partition->lba;
  partition->part_size=0;
  partition->part_type=0;
  /* Test ReiserFS */
  if(hd_read_LBA(disk_car,8, &buffer, hd_offset+128)) /* 64k offset */
    return -1;
  if(!test_rfs(disk_car,(struct reiserfs_super_block*)&buffer,partition)) return 1;
  return 0;
}

int check_part(t_param_disk_cst disk_car,int debug,t_diskext partition)  
{
  int ret=0;
  switch(partition->part_type)
  {
	case P_12FAT:
	case P_16FAT:
	case P_16FATH:
	case P_16FATBD:
	case P_16FATBDH:
	case P_32FAT_1:
	case P_32FAT_2:
	case P_32FAT_1H:
	case P_32FAT_2H:
	  ret=check_FAT(disk_car,partition,debug);
	  break;
  }
  if(ret)
  {
	aff_part_buffer(2,disk_car,partition);
  }
  return ret;
}

