/*

    File: ext2.c

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include "types.h"
#include "common.h"
#include "ext2.h"
#include "interface.h"
#include "fnctdsk.h"

int test_linux(t_param_disk_cst disk_car, struct ext2_super_block *sb,t_diskext partition,const int debug, const int dump_ind)
{
  if(sb->s_magic==EXT2_SUPER_MAGIC)
  {
    if(dump_ind)
    {
      ecrit_rapport("\nLinux magic value at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      /* There is a little offset ... */
      dump(sb,0x200);
    }
    if (sb->s_free_blocks_count >= sb->s_blocks_count) return 1;
    if (sb->s_free_inodes_count >= sb->s_inodes_count) return 1;
    if (sb->s_errors &&
	(sb->s_errors != EXT2_ERRORS_CONTINUE) &&
	(sb->s_errors != EXT2_ERRORS_RO) &&
	(sb->s_errors != EXT2_ERRORS_PANIC))
      return 1;
    if (sb->s_state & ~(EXT2_VALID_FS | EXT2_ERROR_FS))
      return 1;
    if (sb->s_blocks_count == 0) /* reject empty filesystem */
      return 1;
    if(sb->s_log_block_size>2)  /* block size max = 4096 */
      return 1;
    partition->part_size=sb->s_blocks_count*(EXT2_MIN_BLOCK<<sb->s_log_block_size);
    partition->part_type=P_LINUX;
    if(debug)
    {
      ecrit_rapport("Linux [");
//      if(sb->s_volume_name[0])
      ecrit_rapport_string(sb->s_volume_name,16);
      ecrit_rapport("]");
      /* sb->s_last_mounted seems to be unemployed in kernel 2.2.16 */
      if(EXT2_HAS_RO_COMPAT_FEATURE(sb,EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER))
	ecrit_rapport(" - Sparse");
    }
    return 0;
  }
  return 1;
}
