/*

    File: fnctdsk.c

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */


#ifdef DJGPP
#include <curses.h>
#else
#include <ncurses.h>
#include <sys/types.h>
#endif
#include <stdio.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "lang.h"
#include "testdisk.h"
#include "hdaccess.h"
#include "interface.h"
extern char* nom_os[256];

/* BOOT */
dword NT_sect(t_sector_cst B)
{ return *((const dword*)&(B[0x28]));
}

byte NT_clustersize(t_sector_cst B)
{ return B[0x0D];
}

int bootable(t_entree_cst E)
{
  return E[0];
}

dword sect_rel(t_entree_cst E)
{ return *((const dword*)&(E[8]));
}

dword nbr_sect(t_entree_cst E)
{ return *((const dword*)&(E[12]));
}

dword CHS2SR(t_param_disk_cst disk_car,const word C, const byte H, const byte S)
{ return ((dword)C*(disk_car->CHS.head+1)+H)*disk_car->CHS.sector+S;
}

dword CHS2SR2(t_param_disk_cst disk_car,const t_CHS*CHS)
{ return ((dword)CHS->cylinder*(disk_car->CHS.head+1)+CHS->head)*disk_car->CHS.sector+CHS->sector;
}

dword CHS2SRext(t_param_disk_cst disk_car,const uint C, const uint H, const uint S)
{ return ((dword)C*(disk_car->CHS.head+1)+H)*disk_car->CHS.sector+S;
}

dword get_SR(t_param_disk_cst disk_car,t_disk_cst pos)
{ return CHS2SR(disk_car,pos->cylinder, pos->head, pos->sector);
}

dword get_LBA(t_param_disk_cst disk_car,t_disk_cst pos)
{ return get_SR(disk_car,pos)-1;
}

t_disk set_pos(t_param_disk_cst disk_car,t_disk pos, dword sect)
{
  pos->sector=(sect%disk_car->CHS.sector)+1;
  sect/=disk_car->CHS.sector;
  pos->head=sect%(disk_car->CHS.head+1);
  pos->cylinder=sect/(disk_car->CHS.head+1);
  return pos;
}

dword LBA2sector(t_param_disk_cst disk_car, dword sect)
{ return (sect%disk_car->CHS.sector)+1; }

dword LBA2head(t_param_disk_cst disk_car, dword sect)
{ return (sect/disk_car->CHS.sector)%(disk_car->CHS.head+1); }

dword LBA2cylinder(t_param_disk_cst disk_car, dword sect)
{ return (sect/disk_car->CHS.sector)/(disk_car->CHS.head+1); }

void set_pos2(t_param_disk_cst disk_car, t_CHS * CHS, const dword sect)
{
  dword pos=sect-1;
  CHS->sector=(pos%disk_car->CHS.sector)+1;
  pos/=disk_car->CHS.sector;
  CHS->head=pos%(disk_car->CHS.head+1);
  CHS->cylinder=pos/(disk_car->CHS.head+1);
}

dword get_sect_s(t_param_disk_cst disk_car,t_entree_cst E)
{ return CHS2SR(disk_car,s_cyl(E),s_head(E),s_sect(E));
}

dword get_sect_e(t_param_disk_cst disk_car,t_entree_cst E)
{ return CHS2SR(disk_car,e_cyl(E),e_head(E),e_sect(E));
}

void dup_t_CHS(t_CHS * CHS_dest, const t_CHS * CHS_source)
{
  CHS_dest->cylinder=CHS_source->cylinder;
  CHS_dest->head=CHS_source->head;
  CHS_dest->sector=CHS_source->sector;
}

int compare2uint(uint a, uint b)
{
  if(a<b)
    return -1;
  else
    if(a==b)
      return 0;
    else
      return 1;
}

int compare_CHS(const t_CHS *CHS1, const t_CHS *CHS2)
{
  int res;
  res=compare2uint(CHS1->cylinder,CHS2->cylinder);
  if(res!=0)
    return res;
  res=compare2uint(CHS1->head,CHS2->head);
  if(res!=0)
    return res;
  return compare2uint(CHS1->sector,CHS2->sector);
}

void dup_t_param_disk(t_param_disk dest, const t_param_disk src)
{
  dup_t_CHS(&dest->CHS,&src->CHS);
  dest->lba=src->lba;
  dest->disk=src->disk;
#ifdef DJGPP
  dest->mode=src->mode;
  dup_t_CHS(&dest->CHSR,&src->CHSR);
#else
  dest->name=src->name;
#endif
}

void dup_t_diskext(t_diskext dest, t_diskext_cst src)
{
  dest->lba=src->lba;
  dest->disk=src->disk;
  dest->part_type=src->part_type;
  dest->part_size=src->part_size;
  dest->status=src->status;
  dest->order=src->order;
  dest->sect_rel=src->sect_rel;
}

void partition2entree(t_param_disk_cst disk_car,dword pos,t_diskext_cst partition, t_entree *entree)
{
  t_CHS start,end;
  set_pos2(disk_car,&start,partition->lba);
  set_pos2(disk_car,&end,partition->lba+partition->part_size-1);
  if(partition->status==STATUS_PRIM_BOOT)
    (*entree)[0]=0x80;
  else
    (*entree)[0]=0;             /* Non bootable */
  (*entree)[4]=partition->part_type;
  *((dword*)&((*entree)[8]))=partition->lba+1-pos;
  if(start.cylinder>1023)
  {
    (*entree)[1]=254;
    (*entree)[2]=63|((1023>>8)<<6);
    (*entree)[3]=(uchar)1023;
  }
  else
  {
    (*entree)[1]=start.head;
    (*entree)[2]=start.sector|((start.cylinder>>8)<<6);
    (*entree)[3]=start.cylinder;
  }
  if(end.cylinder>1023)
  {
    (*entree)[5]=254;
    (*entree)[6]=63|((1023>>8)<<6);
    (*entree)[7]=(uchar)1023;
  }
  else
  {
    (*entree)[5]=end.head;
    (*entree)[6]=end.sector|((end.cylinder>>8)<<6);
    (*entree)[7]=end.cylinder;
  }
  *((dword*)&((*entree)[12]))=partition->part_size;
}

int entree2partition(t_param_disk_cst disk_car,dword pos,t_diskext partition, t_entree_cst entree, int status)
{
  dword end_sect;
  t_CHS start,end;
  partition->disk=disk_car->disk;
  partition->part_type=os(entree);
  partition->sect_rel=sect_rel(entree);
  partition->lba=pos+partition->sect_rel;
  partition->order=0;
  partition->part_size=nbr_sect(entree);

  start.cylinder=s_cyl(entree);
  start.head=s_head(entree);
  start.sector=s_sect(entree);
  end.cylinder=e_cyl(entree);
  end.head=e_head(entree);
  end.sector=e_sect(entree);

  switch(status)
  {
    case STATUS_PRIM:
      if(partition->part_type==P_EXTENDX || partition->part_type==P_EXTENDED)
	partition->status=STATUS_EXT;
      else
	if(bootable(entree))
	  partition->status=STATUS_PRIM_BOOT;
	else
	  partition->status=status;
      break;
    default:
      partition->status=status;
      break;
  }
  if(start.cylinder<1023)
  {
    if(partition->lba!=get_sect_s(disk_car,entree)-1)
    {
      ecrit_rapport("BAD_RS LBA=%d %d\n",partition->lba,get_sect_s(disk_car,entree)-1);
      return BAD_RS;
    }
  }

  end_sect=partition->lba+1+partition->part_size-1;
  if(end.cylinder<1023)
  {
    if(end_sect!=get_sect_e(disk_car,entree))
      return BAD_SCOUNT;
  }
  /* Check CHS */
  if((start.sector==0)||(start.sector>disk_car->CHS.sector))
    return BAD_SS;
  if((end.sector==0)||(end.sector>disk_car->CHS.sector))
    return BAD_ES;
  if(start.head>disk_car->CHS.head)
    return BAD_SH;
  if(start.cylinder>disk_car->CHS.cylinder)
    return BAD_SC;
  if(end.head>disk_car->CHS.head)
      return BAD_EH;
  if(end.cylinder>disk_car->CHS.cylinder)
    return BAD_EC;
  /* Check start < end */
  if(start.cylinder>end.cylinder)
    return BAD_EBS;
  if(start.cylinder==end.cylinder)
  {
    if(start.head>end.head)
      return BAD_EBS;
    if(start.head==end.head)
      if(start.sector>end.sector)
	return BAD_EBS;
  }
  return 0;
}

const char* errmsg_entree2partition(int errcode)
{
  switch(errcode)
  {
    case BAD_SS: return m_BAD_S_SECT;
    case BAD_ES: return m_BAD_E_SECT;
    case BAD_SH: return m_BAD_S_HEAD;
    case BAD_EH: return m_BAD_E_HEAD;
    case BAD_EBS: return m_END_BFR_START;
    case BAD_RS: return m_BAD_RS;
    case BAD_SC: return m_BAD_S_CYL;
    case BAD_EC: return m_BAD_E_CYL;
    case BAD_SCOUNT: return m_BAD_SCOUNT;
  }
  return "";
}

int read_MBR(t_param_disk_cst disk_car,void *buffer)
{
  if(hd_read_LBA(disk_car,1, buffer, 0))
  {
    doprintf(m_PART_RD_ERR);
    return 1;
  }
  return 0;
}

int write_MBR(t_param_disk_cst disk_car,void *buffer)
{
  if(hd_write_LBA(disk_car,1, buffer, 0))
  {
    doprintf(m_PART_WR_ERR);
    return 1;
  }
  return 0;
}

int write_ext(t_param_disk_cst disk_car,void *buffer, t_diskext_cst partition )
{
  if(hd_write_LBA(disk_car,1, buffer, partition->lba))
  {
    doprintf(m_PART_WR_ERR);
    return 1;
  }
  return 0;
}

t_list_part add_new_partition(t_list_part old, t_list_part new)
{
  if(new==NULL)
    return old;
  new->prev=NULL;
  new->next=old;
  if(old)
    old->prev=new;
  return new;
}

t_list_part insert_new_partition(t_list_part list_part, t_list_part new)
{
  t_list_part prev,next;
  if(!new)
    return list_part;
  if(!list_part || (new->part->lba<list_part->part->lba))
  {
    new->prev=NULL;
    new->next=list_part;
    if(list_part)
      list_part->prev=new;
    return new;
  }
  for(prev=list_part;;prev=next)
  {
    next=prev->next;
    if(!next || (new->part->lba<next->part->lba))
    { /* prev new next */
      prev->next=new;
      new->prev=prev;
      new->next=next;
      if(next)
	next->prev=new;
      return list_part;
    }
  }
}

int check_list_part(t_list_part list_part)
{
  t_list_part prev=NULL,parts;
  ecrit_rapport("\ncheck_list_part\n");
  for(parts=list_part;parts;parts=parts->next)
  {
	ecrit_rapport("%p %p %p\n",parts->prev, parts, parts->next);
	if(prev!=parts->prev)
	{
	  ecrit_rapport("error\n");
	  return 1;
	}
	prev=parts;
  }
  return 0;
}

t_list_part sort_list_part(t_list_part list_part)
{
  t_list_part new_list_part=NULL,parts,next;
  for(parts=list_part;parts;parts=next)
  {
    next=parts->next;
    new_list_part=insert_new_partition(new_list_part,parts);
  }
  return new_list_part;
}
