/*

    File: godmode.c

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#ifdef DJGPP
#include <curses.h>
#else
#include <ncurses.h>
#endif
#include <ctype.h>      /* toupper, tolower */
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "analyse.h"
#include "lang.h"
#include "godmode.h"
#include "testdisk.h"
#include "fat.h"
#include "interface.h"
#include "hdaccess.h"
//#define ESC 0x1B
#define RO 1
#define RW 0
extern char* nom_os[256];
void init_structure(t_param_disk_cst disk_car, t_list_part list_part);
void ask_mbr_order(t_param_disk_cst disk_car,t_list_part list_part);
void init_extended_order(t_list_part list_part);
t_list_part search_part(t_param_disk_cst disk_car,int paranoid,int debug,int dump_ind,int fast_mode,int align);
t_list_part reduce_structure(t_list_part list_part);
t_list_part add_ext_part(t_param_disk_cst disk_car, t_list_part list_part);
int write_mbr(t_param_disk_cst disk_car, t_list_part list_part,int,int);
int write_all_log(t_param_disk_cst disk_car, t_list_part list_part,int,int);
void adjust_part_size(t_param_disk_cst,t_diskext);
int diff(char buffer[0x200], char buffer_org[0x200]);
t_list_part ask_recover_cih(t_param_disk_cst disk_car,t_list_part list_part,int debug);

void init_structure(t_param_disk_cst disk_car, t_list_part list_part)
{
  /* Reconstruit une structure 
   * nbr_prim nombre de partition primaire
   * first_log, last_log : partition etendu du block etendu
   * nbr_log_block : nombre de block de partition etendu
   *                    devrait etre 0 ou 1 */
  int nbr_prim=0, nbr_log_block=0;
  t_list_part first_log=NULL, last_log=NULL, last_prim=NULL;
  t_list_part element;
  for(element=list_part;element;element=element->next)
  {
    if(!can_be_ext(disk_car,element->part))
    {
      nbr_prim++;
      last_prim=element;
      if(last_log && (last_log->next==element))
      { nbr_log_block++; first_log=NULL; last_log=NULL;}
    }
    else
    {
      if(first_log==NULL)
	first_log=element;
      last_log=element;
    }
  }
  /* Verification */
  if(last_log && (last_log->next==NULL))
    nbr_log_block++;
  ecrit_rapport("\nRes: nbr_prim %d, nbr_log_block %d", nbr_prim, nbr_log_block);
  if(nbr_prim+nbr_log_block>4)
  { /* Probleme */
#ifdef DEBUG
    doprintf("\nProbleme");
#endif
  }
  else if(nbr_log_block>1)
  { /* Choix */
#ifdef DEBUG
    doprintf("\nChoix");
#endif
    for(element=list_part;element;element=element->next)
    {
      if(!can_be_ext(disk_car,element->part))
	element->part->status=STATUS_PRIM;
    }
  }
  else
  {
    for(element=list_part;element;element=element->next)
    {
      if(!can_be_ext(disk_car,element->part))
	element->part->status=STATUS_PRIM;
      else
	element->part->status=STATUS_LOG;
    }
  }
  /* The first primary partition is bootable */
  for(element=list_part;element;element=element->next)
    if(element->part->status==STATUS_PRIM)
    {
      element->part->status=STATUS_PRIM_BOOT;
      break;
    }
}

int can_be_ext(t_param_disk_cst disk_car, t_diskext partition)
{
  return((LBA2head(disk_car,partition->lba)>0)&&
      (LBA2cylinder(disk_car,partition->lba)!=0 ||
       LBA2head(disk_car,partition->lba)!=1 ||
       LBA2sector(disk_car,partition->lba)!=1));
}

void only_one_bootable( t_list_part list_part, t_list_part part_boot)
{
  t_list_part element;
  if(part_boot->part->status==STATUS_PRIM_BOOT)
    for(element=list_part;element;element=element->next)
    {
      if((element!=part_boot)&&(element->part->status==STATUS_PRIM_BOOT))
	element->part->status=STATUS_PRIM;
    }
}

int test_structure(t_list_part list_part)
{ /* Return 1 if bad*/
  int nbr_prim=0, nbr_prim_boot=0, nbr_log_block=0;
  t_list_part first_log=NULL;
  t_list_part new_list_part=NULL,element,new_element;
  int res=0;
  for(element=list_part;element;element=element->next)
  {
    switch(element->part->status)
    {
      case STATUS_LOG:
	  if(first_log==NULL)
	  {
	    first_log=element;
	    nbr_log_block++;
	  }
	break;
      case STATUS_PRIM_BOOT:
	if(nbr_prim_boot++)
	  return 1;
      case STATUS_PRIM:
	nbr_prim++;
	first_log=NULL;
	break;
      case STATUS_DELETED:
	break;
    }
  }
  if(nbr_log_block>1 || nbr_log_block+nbr_prim>4)
    return 1;
  /* Sort list_part in new_list_part */
  for(element=list_part;element;element=element->next)
  {
    if(element->part->status!=STATUS_DELETED)
    {
      new_element=(t_list_part)MALLOC(sizeof(*new_element));
//      ecrit_rapport("insert start %d, end %d\n", element->part->lba,element->part->lba+element->part->part_size-1);
      new_element->part=element->part;
      new_element->next=NULL;
      new_element->prev=NULL;
      new_list_part=insert_new_partition(new_list_part,new_element);
    }
  }
  /* Test chevauchement */
  ecrit_rapport("Test chevauchement\n");
  for(element=new_list_part;element;element=element->next)
  {
    ecrit_rapport("start %d, end %d\n", element->part->lba,element->part->lba+element->part->part_size-1);
    if((element->prev && (element->part->lba<=element->prev->part->lba+element->prev->part->part_size-1)) ||
	(element->next && (element->part->lba+element->part->part_size-1>=element->next->part->lba)))
    {
      ecrit_rapport("\nChevauchement\n");
      res=1;
    }
  }
  ecrit_rapport("Test chevauchement fin\n");
  for(element=new_list_part;element;element=new_element)
  {
    new_element=element->next;
    FREE(element);
  }
  return res;
}

t_list_part search_part(t_param_disk_cst disk_car, int paranoid, int debug, int dump_ind,int fast_mode,int align)
{
  dword pos_fin;
  int found;
  int part_offset=PART_OFFSET128;
  int ind_stop=FALSE;
  t_CHS start,end;
  t_list_part list_part=NULL;
  t_diskext partition=(t_diskext)MALLOC(sizeof(*partition));
  partition->disk=disk_car->disk;
  start.cylinder=0;
  start.head=1;
  start.sector=1;
  partition->lba=CHS2SR2(disk_car,&start)-1;
  ecrit_rapport("\nsearch_part()\n");
  /* ncurses: getch will be a non-blocking call */
  while((partition->lba<=disk_car->lba)&&!ind_stop)
  {
    int res;
    move(ANALYSE_Y,ANALYSE_X);
    doprintf("Analyse cylinder %5d/%d",start.cylinder,disk_car->CHS.cylinder);
    refresh();
    end.cylinder=start.cylinder;
    end.head=start.head;
    end.sector=start.sector;
    found=0;
    switch(part_offset)
    {
      case PART_OFFSET128:
	res=search_type_128(disk_car,partition,debug,dump_ind);
	part_offset=PART_OFFSET2;
	break;
      case PART_OFFSET2:
	res=search_type_2(disk_car,partition,debug,dump_ind);
	part_offset=PART_OFFSET1;
	break;
      case PART_OFFSET1:
	res=search_type_1(disk_car,partition,debug,dump_ind);
	part_offset=PART_OFFSET0;
	break;
      case PART_OFFSET0:
	res=search_type_0(disk_car,partition,debug,dump_ind);
	part_offset=PART_QUIT;
	break;
      default:
	ecrit_rapport("\nBUG: unusual value of part_offset\n");
	exit(1);
    }
    if(res<0)
    {
      doprintf(m_READ_ERROR_AT, start.cylinder,start.head,start.sector);
      if(debug)
	ecrit_rapport(m_READ_ERROR_AT, start.cylinder,start.head,start.sector);
    }
    else
    switch(partition->part_type)
    {
      case 0:
      case P_EXTENDED:
      case P_EXTENDX:
	break;
      default:
	{
	  t_diskext new_partition=(t_diskext)MALLOC(sizeof(*new_partition));
	  pos_fin=partition->lba+partition->part_size-1;
	  dup_t_diskext(new_partition,partition);
	  new_partition->status=STATUS_DELETED;
	  new_partition->sect_rel=0;
	  if(debug)
		  aff_part_rapport(1,disk_car,new_partition);
	  if(pos_fin<disk_car->lba)
	  {
	    t_list_part new_list_part=(t_list_part)MALLOC(sizeof(*new_list_part));
	    found=1;
	    end.cylinder=LBA2cylinder(disk_car,pos_fin);
	    end.head=LBA2head(disk_car,pos_fin);
	    end.sector=LBA2sector(disk_car,pos_fin);
	    if(align)
	    {
	      if(end.sector!=disk_car->CHS.sector)
		end.sector=disk_car->CHS.sector;
	      if(end.head!=disk_car->CHS.head)
		end.head=disk_car->CHS.head;
	    }
	    new_partition->part_size=CHS2SR2(disk_car, &end)-1-new_partition->lba+1;
	    if(debug)
	      aff_part_rapport(1,disk_car,new_partition);
	    check_part(disk_car,debug,new_partition);
	    new_list_part->part=new_partition;
	    list_part=insert_new_partition(list_part,new_list_part);
	  }
	  else
	    FREE(new_partition);
	}	
	break;
    }
    if((!paranoid && found) || (part_offset==PART_QUIT))
    {
      part_offset=PART_OFFSET128;
      /* Avance */
      start.sector=1;
      if(found && !paranoid)
      {
	start.cylinder=end.cylinder;
	start.head=end.head;
      }
      if(start.head<disk_car->CHS.head && (!fast_mode || start.head<2))
      {
	start.head++;
      }
      else
      { 
	start.cylinder++;
	start.head=0;
      }
      partition->lba=CHS2SR2(disk_car,&start)-1;
      if(getch_nodelay()=='S')
	ind_stop=1;
    }
  }
  printf("\r                                ");
  FREE(partition);
  return list_part;
}

void ask_mbr_order(t_param_disk_cst disk_car,t_list_part list_part)
{
  t_diskext table[NBR_PART_MAX];
  t_diskext table2[NBR_PART_MAX];
  int nbr_prim=0;
  int i,pos=0;
  int res;
  int car;
  int quit=0;
  t_list_part element;
  /* Initialisation */
  aff_copy();
  move(4,0);
  aff_disk(disk_car);
  mvaddstr(5,0,m_MBR_ORDER);
  mvaddstr(6,0,m_PART_HEADER2);
  for(element=list_part;element;element=element->next)
  {
    if((element->part->status==STATUS_PRIM)||(element->part->status==STATUS_PRIM_BOOT)||(element->part->status==STATUS_EXT))
    {
      table[nbr_prim++]=element->part;
      element->part->order=nbr_prim;
    }
  }
  /* */
  ecrit_rapport("\nSelection des partitions primaires\n");
  for(i=0;i<nbr_prim;i++)
      aff_part_rapport(2,disk_car,table[i]);
  /* */
  do
  {
    uint order;
    /* sort table into table2 */
    int part=0;
    res=0;
    for(order=1;order<=4;order++)
    {
      int nbr=0;
      for(i=0;i<nbr_prim;i++)
	if(table[i]->order==order)
	{
	  table2[part++]=table[i];
	  nbr++;
	}
      res|=(nbr>1);
    }
    if(part!=nbr_prim)
    {
      ecrit_rapport("\nBUG part %d, nbr_prim %d\n", part, nbr_prim);
    }
    /* */
    for(i=0;i<nbr_prim;i++)
    {
      move(5+2+i,0);
      if(i==pos)
	standout();
      aff_part(2,disk_car,table2[i]);
      if(i==pos)
	standend();
    }
    move(20,0);
    if(res)
      doprintf(m_MBR_ORDER_BAD);
    else
      doprintf(m_MBR_ORDER_GOOD);
    refresh();
    car=getch();
    quit=0;
    switch(car)
    {
      case 0x103:
      case 'A':
	if(--pos<0)
	  pos=nbr_prim-1;
	break;
      case 0x102:
      case 'B':
	if(++pos>=nbr_prim)
	  pos=0;
	break;
      case '1':
      case '2':
      case '3':
      case '4':
	table2[pos]->order=car-'0';
	break;
      case 0x105:	/* Right */
      case ' ':
      case 'C':
      case '+':
      case 'l':
	if(++table2[pos]->order>4)
	  table2[pos]->order=1;
	break;
      case 0x104:	/* Left arrow */
      case 'D':
      case '-':
      case 'S':
	if(--table2[pos]->order<1)
	  table2[pos]->order=4;
	break;
      case 'q':
      case '\r':
      case '\n':
      case 'M':
	quit=1;
	break;
    }
    refresh();
  } while(res || !quit);
}

void init_extended_order(t_list_part list_part)
{ /* Initialisation */
  int nbr_log=0;
  t_list_part element;
  for(element=list_part;element;element=element->next)
  {
    if(element->part->status==STATUS_LOG)
      element->part->order=(++nbr_log)+4;
  }
}

t_list_part reduce_structure(t_list_part list_part)
{ 
  t_list_part element=list_part;
  t_list_part prev=NULL;
  while(element)
  {
    t_list_part next=element->next;
    if(element->part->status==STATUS_DELETED)
    {
      if(prev==NULL)
	list_part=next;
      else
	prev->next=next;
      if(next)
	next->prev=prev;
      FREE(element->part);
      FREE(element);
    }
    else
      prev=element;
    element=next;
  }
  return list_part;
}

t_list_part add_ext_part(t_param_disk_cst disk_car, t_list_part list_part)
{
  /* list_part need to be sorted! */
  /* All extended partitions of an P_EXTENDX  are P_EXTENDED */
  t_list_part element,deb=NULL,fin=NULL;
  int nbr_entree=0;
  t_CHS start,end;
  t_list_part new_list_part;
  t_diskext new_partition;
  for(element=list_part;element;element=element->next)
  {
    if(element->part->status==STATUS_LOG)
    {
      if(deb==NULL)
      {
	deb=element;
	nbr_entree++;
      }
      fin=element;
    }
    else
      nbr_entree++;
  }
  if(!deb)
    return list_part;
  new_partition=(t_diskext)MALLOC(sizeof(*new_partition));
  new_partition->disk=deb->part->disk;
  new_partition->sect_rel=0;
  start.sector=1;
  if(nbr_entree==4)
  {
    start.cylinder=(deb->prev==NULL?1:LBA2cylinder(disk_car,deb->prev->part->lba+deb->prev->part->part_size-1)+1);
    start.head=0;
    end.cylinder=(fin->next==NULL?disk_car->CHS.cylinder:LBA2cylinder(disk_car,fin->next->part->lba+fin->next->part->part_size-1)-1);
    end.head=disk_car->CHS.head;
    end.sector=disk_car->CHS.sector;
  }
  else
  {
    start.cylinder=LBA2cylinder(disk_car,deb->part->lba);
    if(LBA2head(disk_car,deb->part->lba)==0)
      ecrit_rapport("BUG while creating a minimal extended partition\n");
    start.head=LBA2head(disk_car,deb->part->lba)-1;
    end.cylinder=LBA2cylinder(disk_car,fin->part->lba+fin->part->part_size-1);
    end.head=disk_car->CHS.head;
    end.sector=disk_car->CHS.sector;
  }
  new_partition->part_type=(end.cylinder>1023?P_EXTENDX:P_EXTENDED);
  new_partition->status=STATUS_EXT;
  new_partition->lba=CHS2SR2(disk_car,&start)-1;
  new_partition->part_size=CHS2SR2(disk_car,&end)-new_partition->lba;
  new_list_part=(t_list_part)MALLOC(sizeof(*new_list_part));
  new_list_part->part=new_partition;
  new_list_part->prev=new_list_part->next=NULL;
  return insert_new_partition(list_part,new_list_part);
}

int write_mbr(t_param_disk_cst disk_car, t_list_part list_part, int ro,int debug)
{
  t_list_part element;
  uchar buffer[0x200];
  uchar buffer_org[0x200];
  if(debug)
	ecrit_rapport("\nwrite_mbr: starting...");
  if(read_MBR(disk_car,&buffer_org)) return 1;
  memset((void *)&buffer,0,0x200);
  memcpy(buffer,buffer_org,TAB_PART);
  buffer[0x1FE]=0x55;
  buffer[0x1FF]=0xAA;
  for(element=list_part;element;element=element->next)
  {
    switch(element->part->status)
    {
      case STATUS_PRIM:
      case STATUS_PRIM_BOOT:
      case STATUS_EXT:
	if((element->part->order>=1) && (element->part->order<=4))
	  partition2entree(disk_car,1,element->part, 
	    (t_entree*)&buffer[TAB_PART+(element->part->order-1)*0x10]);
	break;
    }
  }
  if(debug)
  {
    int i;
    for(i=0;i<4;i++)
      aff_entree(&buffer[0x10*i+TAB_PART]);
#ifdef DUMP
    diff(buffer,buffer_org);
#endif
  }
  if(!ro)
    return write_MBR(disk_car,&buffer);
  return 0;
}

#ifdef DUMP
int diff(char buffer[0x200], char buffer_org[0x200])
{
  int j;
  if(memcmp(buffer,buffer_org,0x200))
  {
    for(j=0;j<4;j++)
      if(os(&buffer[0x10*j+TAB_PART]))
	aff_entree(&buffer_org[0x10*j+TAB_PART]);
    printf("\nSectors are differents\n");
    for(j=0;j<0x200;j++)
      if(buffer[j]!=buffer_org[j])
	printf("%02X ", j);
  }
  return 0;
}
#endif

int write_all_log(t_param_disk_cst disk_car, t_list_part list_part, int ro, int debug)
{
  t_list_part element;
  int i;
  t_list_part pos_ext=NULL;
  t_list_part table[NBR_PART_MAX];
  int nbr_log=0;
  dword pos;
  t_diskext bloc_ext,bloc_nextext;
  if(debug)
	ecrit_rapport("write_all_log: starting...\n");
  for(element=list_part;element;element=element->next)
  {
    switch(element->part->status)
	{
	  case STATUS_EXT:          /* Pas bootable */
		if((element->part->part_type==P_EXTENDED) ||
			(element->part->part_type==P_EXTENDX))
		  pos_ext=element;
		break;
	  case STATUS_LOG:
		table[nbr_log++]=element;
		break;
	}
  }
  if(pos_ext==NULL)
  {
    doprintf(m_NO_EXT_PART);
    return 1;
  }
  bloc_ext=(t_diskext)MALLOC(sizeof(*bloc_ext));
  bloc_nextext=(t_diskext)MALLOC(sizeof(*bloc_nextext));
  dup_t_diskext(bloc_ext,pos_ext->part);
  pos=pos_ext->part->lba-1;
  for(i=0;i<nbr_log;i++)
  {
    /* tabext: partition etendu conteneur */
    /* tab[part[table[i]]: partition */
    int j;
    uchar buffer[0x200];
    uchar buffer_org[0x200];
    if(debug)
    {
      ecrit_rapport("write_all_log: CHS: %d %d %d\n", LBA2cylinder(disk_car,bloc_ext->lba), LBA2head(disk_car,bloc_ext->lba), LBA2sector(disk_car,bloc_ext->lba));
    }
    if(hd_read_LBA(disk_car,1, &buffer_org, bloc_ext->lba))
      return 1;
    memset(buffer,0,0x200);
    memcpy(buffer,buffer_org,TAB_PART);
    buffer[0x1FE]=0x55;
    buffer[0x1FF]=0xAA;
    partition2entree(disk_car,bloc_ext->lba-1,table[i]->part, (t_entree*)&buffer[TAB_PART+0x0]);
    if(i<nbr_log-1)
    { /* Construit le pointeur vers la prochaine partition logique */
      t_CHS nextext_start,nextext_end;
      bloc_nextext->part_type=P_EXTENDED;       /* Never P_EXTENDX */
      bloc_nextext->disk=table[i+1]->part->disk;
      nextext_start.cylinder=LBA2cylinder(disk_car,table[i+1]->part->lba);
      nextext_start.head=LBA2head(disk_car,table[i+1]->part->lba)-1;
      nextext_start.sector=1;
      bloc_nextext->lba=CHS2SR2(disk_car,&nextext_start)-1;

      nextext_end.cylinder=LBA2cylinder(disk_car,table[i+1]->part->lba+table[i+1]->part->part_size-1);
      nextext_end.head=LBA2head(disk_car,table[i+1]->part->lba+table[i+1]->part->part_size-1);
      nextext_end.sector=LBA2sector(disk_car,table[i+1]->part->lba+table[i+1]->part->part_size-1);
      bloc_nextext->part_size=CHS2SR2(disk_car,&nextext_end)-1-bloc_nextext->lba+1;
      partition2entree(disk_car,pos,bloc_nextext, (t_entree*)&buffer[TAB_PART+0x10]);
    }
    if(ro)
    {
      for(j=0;j<4;j++)
	if(os(&buffer[0x10*j+TAB_PART]))
	  aff_entree_rapport(&buffer[0x10*j+TAB_PART]);
#ifdef DUMP
      diff(buffer,buffer_org);
#endif
    }
    else
      write_ext(disk_car,&buffer_org,bloc_ext);
    memcpy(bloc_ext,bloc_nextext,sizeof(struct cellule_diskext));
  }
  FREE(bloc_ext);
  FREE(bloc_nextext);
  return 0;
}

t_list_part ask_recover_cih(t_param_disk_cst disk_car,t_list_part list_part,int debug)
{ /* tab_part need to be sorted */
  if((list_part==NULL) || (LBA2cylinder(disk_car,list_part->part->lba)>0))
  {
    t_CHS start,end;
    t_diskext partition=(t_diskext)MALLOC(sizeof(*partition));
    partition->disk=disk_car->disk;
    partition->part_type=0;
    start.cylinder=0;
    start.head=1;
    start.sector=1;
    partition->lba=CHS2SR2(disk_car,&start)-1;
    if(list_part==NULL)
      end.cylinder=disk_car->CHS.cylinder;
    else
      end.cylinder=LBA2cylinder(disk_car,list_part->part->lba)-1;
    end.head=disk_car->CHS.head;
    end.sector=disk_car->CHS.sector;
    partition->part_size=CHS2SR2(disk_car,&end)-1-partition->lba+1;
    if(recover_FAT(disk_car,partition,debug)==0)
    {
      t_list_part new_list_part=(t_list_part)MALLOC(sizeof(*new_list_part));
      new_list_part->part=partition;
      new_list_part->prev=new_list_part->next=NULL;
      return insert_new_partition(list_part,new_list_part);
    }
    FREE(partition);
  }
  return list_part;
}


int god_mode(t_param_disk_cst disk_car,int paranoid,int debug,int dump_ind, int fast_mode,int align)
{
  t_list_part list_part;
  t_list_part element;
  aff_disk_rapport(disk_car);
  list_part=search_part(disk_car,paranoid,debug,dump_ind,fast_mode,align);
  if(!list_part)
    return 1;   /* Err de lecture */
//  list_part=ask_recover_cih(list_part,debug);
  init_structure(disk_car,list_part);
#ifdef DEBUG
  check_list_part(list_part);
#endif
  /* Ajoute dans le fichier de log les partitions trouvees */
  ecrit_rapport("\nResults");
  for(element=list_part;element;element=element->next)
    aff_part_rapport(1,disk_car,element->part);
  doprintf(m_GOD_MODE);
  ask_structure(disk_car,list_part);
  list_part=reduce_structure(list_part);
#ifdef DEBUG
  check_list_part(list_part);
#endif
  /* Sort list_part */
  list_part=sort_list_part(list_part);
#ifdef DEBUG
  check_list_part(list_part);
#endif
  /* Creer la partition etendue globale */
  list_part=add_ext_part(disk_car,list_part);
  ecrit_rapport("\nResults with extended");
  for(element=list_part;element;element=element->next)
    aff_part_rapport(1,disk_car,element->part);
  /* Demande l'ordre des entrees dans le MBR */
  ask_mbr_order(disk_car,list_part);
  /* Demande l'ordre des partitions etendues */
  init_extended_order(list_part);
  if(interface_write(disk_car,list_part))
  {
    write_mbr(disk_car,list_part,RW,debug);
    write_all_log(disk_car,list_part,RW,debug);
  }
  else
  {
    write_all_log(disk_car,list_part,RO,debug);
  }
  /* Libere la memoire */
  if(debug)
    ecrit_rapport("god_mode: free the memory\n");
  element=list_part;
  while(element)
  {
    t_list_part next=element->next;
    FREE(element->part);
    FREE(element);
    element=next;
  }
  return 0;
}
