/*

    File: hdaccess.c

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#define LINUX_WRITE 1
#ifdef LINUX_WRITE
#define open_mode O_RDWR
#else
#define open_mode O_RDONLY
#endif
#ifdef DJGPP
#include <stdlib.h>     /* atexit */
#include <go32.h>       /* dosmemget/put */
#include <dpmi.h>
#include <bios.h>       /* bios_k* */
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "hdaccess.h"
#include "interface.h"
#define HD_RW_BUF_SIZ 0x10
#define HDPARM_BUF_SIZ 0x1A
#define MAX_IO_NBR 10
#define PARA_CMD         8
int hd_read(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, t_disk_cst lieu);
int hd_super(int cmd, t_param_disk_cst disk_car, loff_t hd_offset, int nsects, void *buffer);
int hd_identify_bios(t_param_disk param_disk,const int debug);
int hd_identify_enh_bios(t_param_disk param_disk,const int debug);
int check_enh_bios(t_param_disk param_disk, const int debug);
static int dos_segment = 0;
static int dos_selector = 0;

void free_dos_buffer(void)
{
  __dpmi_free_dos_memory(dos_selector);
  dos_segment = dos_selector = 0;
}

int alloc_dos_buffer(void)
{
  if (dos_segment)
    return 0;
  if ((dos_segment = __dpmi_allocate_dos_memory(18*512/16, &dos_selector))== -1)
  {
    dos_segment = 0;
    return 1;
  }
  atexit(free_dos_buffer);
  return 0;
}

int hd_super(int cmd, t_param_disk_cst disk_car, loff_t hd_offset, int nsects, void *buffer)
{ 
  __dpmi_regs r;
  uchar buf[HD_RW_BUF_SIZ];
  int dos_segment2, dos_selector2, xfer2=nsects*512;
  if(disk_car->mode==0)
  { // Limite CHS= 1023,255,63 = 8,064Mo ~= 7.8 Go
    int head, track, sector;
    sector=(hd_offset%disk_car->CHSR.sector)+1;
    hd_offset/=disk_car->CHSR.sector;
    head=hd_offset%(disk_car->CHSR.head+1);
    track=hd_offset/(disk_car->CHSR.head+1);
    if(track<1024)
      return biosdisk(cmd, disk_car->disk, head, track, sector, nsects, buffer);
    return 1;
  }
  if(dos_segment==0)
    if(alloc_dos_buffer())
      return 1;
  if ( (dos_segment2=__dpmi_allocate_dos_memory((xfer2 + 15) >> 4, &dos_selector2)) == -1 )
    return 1;
  *(word*)&buf[0]=HD_RW_BUF_SIZ;
  *(word*)&buf[2]=nsects;
  *(dword*)&buf[0x4]=dos_segment2<<16;
  *(dword*)&buf[0x8]=hd_offset;
  *(dword*)&buf[0xC]=0;

  r.x.ds = dos_segment;
  r.x.si = 0;
  r.h.dl = disk_car->disk;
  switch(cmd)
  {
    case 2:
      r.h.ah = 0x42;
      break;
    case 3:
      r.x.ax = 0x4300;
      dosmemput(buffer,xfer2,dos_segment2<<4);
      break;
	case 0x0C:
	  r.h.ah = 0x47;
	  break;
  }
  dosmemput(&buf, HD_RW_BUF_SIZ, dos_segment<<4);
  __dpmi_int(0x13, &r);
  if(cmd==2)
    dosmemget(dos_segment2<<4, xfer2, buffer);
  __dpmi_free_dos_memory(dos_selector2);
#ifdef DEBUG
  if(r.h.ah)
  {
    doprintf(" %ld err %02X %04X\n",*(dword*)&buf[0x8], r.h.ah,r.x.flags);
  }
#endif
  return r.h.ah;
}

int check_enh_bios(t_param_disk param_disk, const int debug)
{
  __dpmi_regs r;
  r.h.ah = 0x41;
  r.x.bx = 0x55AA;
  r.h.dl = param_disk->disk;
  __dpmi_int(0x13, &r);
  if(r.x.bx != 0xAA55)  /* INT 13 Extensions not installed */
  {
    if(debug)
      ecrit_rapport("Disk %02X - INT 13 Extensions not installed\n",param_disk->disk);
    return 0;
  }
  if(debug)
  {
    ecrit_rapport("Disk %02X ",param_disk->disk);
    switch(r.h.ah)
	{
	  case 0x01:
		ecrit_rapport("Enhanced BIOS 1.x");
		break;
	  case 0x20:
		ecrit_rapport("Enhanced BIOS 2.0 / EDD-1.0");
		break;
	  case 0x21:
		ecrit_rapport("Enhanced BIOS 2.1 / EDD-1.1");
		break;
	  case 0x30:
		ecrit_rapport("Enhanced BIOS EDD-3.0");
		break;
	  default:
		ecrit_rapport("Enhanced BIOS unknow %02X",r.h.ah);
		break;
	}
    if((r.x.cx & 1)!=0)
      ecrit_rapport(" - R/W/I");
    if((r.x.cx & 4)!=0)
      ecrit_rapport(" - Identify");
    ecrit_rapport("\n");
  }
  return ((r.x.cx&1)!=0);
}

int hd_identify_enh_bios(t_param_disk param_disk,const int debug)
{
  t_disk pos;
  __dpmi_regs r;
  uchar buf[0x200];	/* Don't change it! */
  buf[0]=HDPARM_BUF_SIZ;
  buf[1]=0;
  if(dos_segment==0)
    if(alloc_dos_buffer())
      return 1;
  r.h.ah = 0x48;
  r.x.ds = dos_segment;
  r.x.si = 0;
  r.h.dl = param_disk->disk;
  dosmemput(&buf, HDPARM_BUF_SIZ, dos_segment<<4);
  __dpmi_int(0x13, &r);
  dosmemget(dos_segment<<4, HDPARM_BUF_SIZ, &buf);
  if(r.h.ah)
    return 1;
  param_disk->CHS.cylinder=*(word*)&buf[0x04];
  param_disk->CHS.head=*(word*)&buf[0x08];
  param_disk->CHS.sector=*(word*)&buf[0x0C];
  param_disk->lba=(*(dword*)&buf[0x10]);
  if(param_disk->CHS.cylinder)
  {
	param_disk->CHS.cylinder--;
	param_disk->CHS.head--;
/* */
  }
  else
  {   /* ATTENTION qword en fait */
	if(debug)
	{
	  ecrit_rapport("Disk %02X Computes CHS from number of sectors - Data CHS=(%d,%d,%d) T=%ld\n",param_disk->disk, param_disk->CHS.cylinder,param_disk->CHS.head,param_disk->CHS.sector,param_disk->lba);
	}
	param_disk->CHS.cylinder=(param_disk->lba/255)/63-1;
	param_disk->CHS.head=255-1;
	param_disk->CHS.sector=63;
  }
  ecrit_rapport("lba %ld %ld",param_disk->lba,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector);
  if(param_disk->lba==0)
  {
    param_disk->lba=(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector;
  }
  pos=(t_disk)MALLOC(sizeof(*pos));
  pos->disk=param_disk->disk;
  pos->cylinder=param_disk->CHS.cylinder+1;
  pos->head=param_disk->CHS.head;
  pos->sector=param_disk->CHS.sector;
  if(hd_read(param_disk,1, &buf, pos)==0)
  {
    param_disk->CHS.cylinder++;
    if(debug)
      ecrit_rapport("Disk %02X hd_identify_enh_bios, BIOS hides the last cylinder.\n",param_disk->disk);
  }
  FREE(pos);
  param_disk->CHSR.cylinder=param_disk->CHS.cylinder;
  param_disk->CHSR.head=param_disk->CHS.head;
  param_disk->CHSR.sector=param_disk->CHS.sector;
  if(debug)
  {
    aff_disk_rapport(param_disk);
  }
  return 0;
}

int hd_identify_bios(t_param_disk param_disk,const int debug)
{
  uchar buf[0x200];
  if(biosdisk(PARA_CMD,param_disk->disk,0,0,1,1,buf))
    return 1;
  param_disk->CHS.cylinder=((buf[0] & 0x0C0)<<2)|buf[1];
  param_disk->CHS.head=buf[3];
  param_disk->CHS.sector=buf[0] & 0x3F;
  if((param_disk->CHS.cylinder<1023)&& (!biosdisk(0x0C, param_disk->disk, 0, param_disk->CHS.cylinder+1, param_disk->CHS.head, param_disk->CHS.sector, NULL)))
  {
    param_disk->CHS.cylinder++;
    if(debug)
      ecrit_rapport("Disk %02X hd_identify_bios, BIOS hides the last cylinder.\n",param_disk->disk);
  }
  param_disk->lba=(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector;
  param_disk->CHSR.cylinder=param_disk->CHS.cylinder;
  param_disk->CHSR.head=param_disk->CHS.head;
  param_disk->CHSR.sector=param_disk->CHS.sector;
  if(debug)
  {
    aff_disk_rapport(param_disk);
  }
  return 0;
}

int hd_read_LBA(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, loff_t hd_offset)
{
  int i,rc=-1;
  for(i=0;(rc!=0) && (i<MAX_IO_NBR);i++)
	rc=hd_super(2, disk_car, hd_offset, nbr_secteur, nom_buffer);
  return rc;
}

int hd_read(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, t_disk_cst lieu)
{
  return hd_read_LBA(disk_car,nbr_secteur,nom_buffer, get_LBA(disk_car,lieu));
}


int hd_write_LBA(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, loff_t hd_offset)
{
  int i,rc=-1;
  for(i=0;(rc!=0)&&(i<MAX_IO_NBR);i++)
    rc=hd_super(3, disk_car, hd_offset, nbr_secteur, nom_buffer);
  return rc;
}

int hd_parse(t_param_disk param_disk[NBR_DISK_MAX],const int debug)
{
  int i;
  int nbr_disk=0;
  for(i=0x80;i<0x88;i++)
  {
    /* standard BIOS access */
    param_disk[nbr_disk]->disk=i;
    param_disk[nbr_disk]->mode=0;
    if(!hd_identify_bios(param_disk[nbr_disk],debug))
    {
      nbr_disk++;
      /* enhanced BIOS access */
      param_disk[nbr_disk]->disk=i;
      param_disk[nbr_disk]->mode=1;
      if(check_enh_bios(param_disk[nbr_disk],debug))
      {
        if(!hd_identify_enh_bios(param_disk[nbr_disk],debug))
	{
	  if((param_disk[nbr_disk]->CHS.head==param_disk[nbr_disk-1]->CHS.head)
	      &&(param_disk[nbr_disk]->CHS.sector==param_disk[nbr_disk-1]->CHS.sector))
	  {
	    /* Remove unnecessary standard disk access */
	    dup_t_param_disk(param_disk[nbr_disk-1],param_disk[nbr_disk]);
	  }
	  else
	  {
	    /* Mixed mode */
	    param_disk[nbr_disk-1]->mode=2; /* extended functions */
	    /* standard geometry H,S, compute C from lba */
	    param_disk[nbr_disk-1]->lba=param_disk[nbr_disk]->lba;
	    param_disk[nbr_disk-1]->CHS.cylinder=(param_disk[nbr_disk-1]->lba/(param_disk[nbr_disk-1]->CHS.head+1))/param_disk[nbr_disk-1]->CHS.sector-1;
//	    nbr_disk++;
	  }
	}
      }
    }
  }
  return nbr_disk;
}
#elif defined(LINUX)
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/hdreg.h>
#include <fcntl.h> 
#include <sys/mount.h>
#include <string.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "hdaccess.h"

loff_t lseek64(int,loff_t, int);
/* if lseek64 don't work well, try lseek64 */

int hd_read_LBA(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, loff_t pos)
{
  int i;
//  ecrit_rapport("hd_read_LBA(%d,%d,buffer,%lx)\n", disk_car->disk,nbr_secteur,pos);
  i=lseek64(disk_car->disk,pos<<9,SEEK_SET);
  if(i==-1)
  {
    ecrit_rapport("hd_read_LBA(%d,%d,buffer,%ld) seek err\n", disk_car->disk,nbr_secteur,pos);
    return -1;
  }
  if(read(disk_car->disk, nom_buffer, nbr_secteur*0x200)==-1)
  {
    ecrit_rapport("hd_read_LBA(%d,%d,buffer,%ld) read err\n", disk_car->disk,nbr_secteur,pos);
	return -1;
  }
  return 0;
}

int hd_write_LBA(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, loff_t pos)
{
#ifdef LINUX_WRITE
  int i;
  i=lseek64(disk_car->disk,pos<<9,SEEK_SET);
  if(i==-1)
    return -1;
  if(write(disk_car->disk, nom_buffer, nbr_secteur*0x200)<0)
    return -1;
  return 0;
#else
  return 1;
#endif
}

void test_disk_availability(int * nbr_disk,t_param_disk disk_car, char *device);
void test_disk_availability(int * nbr_disk,t_param_disk disk_car, char *device)
{
  struct hd_geometry geometry;
  int hd_h = open(device, open_mode);
  if(hd_h>0)
  {
    /* Little trick from Linux fdisk */
    /* Blocks are visible in more than one way:
       e.g. as block on /dev/hda and as block on /dev/hda3
       By a bug in the Linux buffer cache, we will see the old
       contents of /dev/hda when the change was made to /dev/hda3.
       In order to avoid this, discard all blocks on /dev/hda. */
    ioctl(hd_h, BLKFLSBUF);	/* ignore errors */
    				/* e.g. Permission Denied */
    if (!ioctl(hd_h, HDIO_GETGEO, &geometry))
    { /* I can get the geometry */
      disk_car->name=MALLOC(strlen(device)+1);
      strcpy(disk_car->name,device);
      disk_car->lba=geometry.cylinders*geometry.heads*geometry.sectors;
      disk_car->CHS.cylinder= geometry.cylinders-1;
      disk_car->CHS.head=geometry.heads-1;
      disk_car->CHS.sector= geometry.sectors;
      disk_car->disk= hd_h;
      (*nbr_disk)++;
    }
    else
      close(hd_h);
  }
}

int hd_parse(t_param_disk param_disk[NBR_DISK_MAX],const int debug)
{
  int i;
  char device_ide[]="/dev/hda";
  char device_scsi[]="/dev/sda";
  char device_ida[]="/dev/ida/c0d0";
  int nbr_disk=0;
  /* Disk IDE */
  for(i=0;i<8;i++)
  {
    device_ide[strlen(device_ide)-1]='a'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_ide);
  }
  /* Disk SCSI */
  for(i=0;i<8;i++)
  {
    device_scsi[strlen(device_scsi)-1]='a'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_scsi);
  }
  /* Device RAID Compaq */
  test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_ida);
  return nbr_disk;
}
#else
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h> 
#include <sys/disklabel.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "hdaccess.h"
int hd_read_LBA(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, loff_t pos)
{
  int i;
  i=lseek(disk_car->disk,pos<<9,SEEK_SET);
  if(i==-1)
  {
    ecrit_rapport("hd_read_LBA(%d,%d,buffer,%ld) seek err\n", disk_car->disk,nbr_secteur,pos);
    return -1;
  }
  if(read(disk_car->disk, nom_buffer, nbr_secteur*0x200)==-1)
  {
    ecrit_rapport("hd_read_LBA(%d,%d,buffer,%ld) read err\n", disk_car->disk,nbr_secteur,pos);
	return -1;
  }
  return 0;
}

int hd_write_LBA(t_param_disk_cst disk_car,const uint nbr_secteur, void *nom_buffer, loff_t pos)
{
#ifdef LINUX_WRITE
  int i;
  i=lseek(disk_car->disk,pos<<9,SEEK_SET);
  if(i==-1)
    return -1;
  if(write(disk_car->disk, nom_buffer, nbr_secteur*0x200)<0)
    return -1;
  return 0;
#else
  return 1;
#endif
}

void test_disk_availability(int * nbr_disk,t_param_disk disk_car, char *device);
void test_disk_availability(int * nbr_disk,t_param_disk disk_car, char *device)
{
  struct disklabel geometry;
  int hd_h = open(device, open_mode);
  if(hd_h>0)
  {
    if (!ioctl(hd_h, DIOCGDINFO, &geometry))
    { /* I can get the geometry */
      disk_car->name=MALLOC(strlen(device)+1);
      strcpy(disk_car->name,device);
/*
      disk_car->CHS.cylinder= geometry.cylinders-1;
      disk_car->CHS.head=geometry.heads-1;
      disk_car->CHS.sector= geometry.sectors;
 */
      disk_car->lba=geometry.d_ncylinders*geometry.d_ntracks*geometry.d_nsectors;
      disk_car->CHS.cylinder=geometry.d_ncylinders-1;
      disk_car->CHS.head=geometry.d_ntracks-1;
      disk_car->CHS.sector=geometry.d_nsectors;
      disk_car->disk= hd_h;
      (*nbr_disk)++;
    }
    else
      close(hd_h);
  }
}

int hd_parse(t_param_disk param_disk[NBR_DISK_MAX],const int debug)
{
  int i;
  char device_ide[]="/dev/rwd0";
  char device_scsi[]="/dev/rda0";
  char device_optdisk[]="/dev/rod0";
  int nbr_disk=0;
  /* Disk IDE */
  for(i=0;i<8;i++)
  {
    device_ide[strlen(device_ide)-1]='0'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_ide);
  }
  /* Disk SCSI */
  for(i=0;i<8;i++)
  {
    device_scsi[strlen(device_scsi)-1]='0'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_scsi);
  }
  /* optical disks */
  for(i=0;i<8;i++)
  {
    device_optdisk[strlen(device_scsi)-1]='a'+i;
    test_disk_availability(&nbr_disk,param_disk[nbr_disk], device_optdisk);
  } 
  return nbr_disk;
}
#endif
