#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef DJGPP
#include <curses.h>
#else
#include <ncurses.h>
#endif
#include <string.h>
#include <ctype.h>
#include "types.h"
#include "common.h"
#include "lang.h"
#include "interface.h"
#include "godmode.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "lang.h"
#include "hdaccess.h"

extern char* nom_os[256];
int get_string(char *str, int len, char *def);
void interface_analyse(t_param_disk_cst param_disk, const int debug);
void interface_recovery(t_param_disk_cst param_disk,const int paranoid, const int debug, const int dump_ind, const int fast_mode, const int align);
void interface_options(int *paranoid,int *dump_ind,int *fast_mode,int *align);
void interface_editor_position(t_param_disk_cst param_disk, t_disk position);
void interface_editor(t_param_disk_cst param_disk);
#define GS_DEFAULT -1
#define GS_ESCAPE -2
int dump2(const uchar *nom_dump,uint lng);
void interface_editor2(t_param_disk_cst param_disk);
void aff_CHS_from_LBA(t_param_disk_cst disk_car, dword hd_offset);
void aff_CHS_from_LBA_buffer(t_param_disk_cst disk_car, dword hd_offset);
void aff_CHS_from_LBA_rapport(t_param_disk_cst disk_car, dword hd_offset);


int get_string(char *str, int len, char *def)
{
    char c;
    int i = 0;
    int x, y;
    int use_def = FALSE;

    getyx(stdscr, y, x);
    clrtoeol();

    str[0] = 0;

    if (def != NULL) {
	mvaddstr(y, x, def);
	move(y, x);
	use_def = TRUE;
    }

    refresh();
    while ((c = getch()) != '\n' && c != CR) {
	switch (c) {
	case ESC:
	    move(y, x);
	    clrtoeol();
	    refresh();
	    return GS_ESCAPE;
	case DEL:
	case '\b':
	    if (i > 0) {
		str[--i] = 0;
		mvaddch(y, x+i, ' ');
		move(y, x+i);
	    } else if (use_def) {
		clrtoeol();
		use_def = FALSE;
	    } else
		putchar(BELL);
	    break;
	default:
	    if (i < len && isprint(c)) {
		mvaddch(y, x+i, c);
		if (use_def) {
		    clrtoeol();
		    use_def = FALSE;
		}
		str[i++] = c;
		str[i] = 0;
	    } else
		putchar(BELL);
	}
	refresh();
    }

    if (use_def)
	return GS_DEFAULT;
    else
	return i;
}

int getch_nodelay()
{
  int res;
  nodelay(stdscr,TRUE);
  res=toupper(getch());
  nodelay(stdscr,FALSE);
  return res;
}

/*
 * Actual function which prints the button bar and highlights the active button
 * Should not be called directly. Call function menuSelect instead.
 */

int menuUpdate( int y, int x, const struct MenuItem *menuItems, const int itemLength, const char *available, const int menuType, int current )
{
    int i, lmargin = x, ymargin = y;
    const char *mcd;

    /* Print available buttons */
    for( i = 0; menuItems[i].key; i++ )
    {
      char buff[80];
      int lenName;
      const char *mi;
      move( y, x ); clrtoeol();

	/* Search next available button */
	while( menuItems[i].key && !strchr(available, menuItems[i].key) )
	{
	    i++;
	}
	if( !menuItems[i].key ) break; /* No more menu items */

	/* If selected item is not available and we have bypassed it,
	   make current item selected */
	if( current < i && menuItems[current].key < 0 ) current = i;

	/* If current item is selected, highlight it */
	if( current == i ) /*attron( A_REVERSE )*/ standout ();

	/* Print item */
	mi = menuItems[i].name;
	lenName = strlen( mi );
#ifdef DJGPP
	if(lenName>=sizeof(buff))
	{
	  ecrit_rapport("\nBUG: %s\n",mi);
	}
	if(lenName > itemLength)
	{
	  if( menuType & MENU_BUTTON )
		sprintf(buff, "[%s]",mi);
	  else
		sprintf(buff, "%s",mi);
	}
	else
	{
	  if( menuType & MENU_BUTTON )
		sprintf( buff, "[%*s%-*s]", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	  else
		sprintf( buff, "%*s%-*s", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	}
#else
	if(lenName>=sizeof(buff))
	{
	  ecrit_rapport("\nBUG: %s\n",mi);
	}
	if(lenName > itemLength)
	{
	  if( menuType & MENU_BUTTON )
		snprintf(buff, sizeof(buff),"[%s]",mi);
	  else
		snprintf(buff, sizeof(buff),"%s",mi);
	}
	else
	{
	  if( menuType & MENU_BUTTON )
		snprintf( buff, sizeof(buff),"[%*s%-*s]", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	  else
		snprintf( buff, sizeof(buff),"%*s%-*s", (itemLength - lenName) / 2, "",
			(itemLength - lenName + 1) / 2 + lenName, mi );
	}
#endif
	mvaddstr( y, x, buff );

	/* Lowlight after selected item */
	if( current == i ) /*attroff( A_REVERSE )*/ standend ();

	/* Calculate position for the next item */
	if( menuType & MENU_VERT )
	{
	    y += 1;
	    if( y >= WARNING_START )
	    {
		y = ymargin;
		x += itemLength + MENU_SPACING;
		if( menuType & MENU_BUTTON ) x += 2;
	    }
	}
	else
	{
	    x += itemLength + MENU_SPACING;
	    if( menuType & MENU_BUTTON ) x += 2;
	    if( x > COLUMNS - lmargin - 12 )
	    {
		x = lmargin;
		y ++ ;
	    }
	}
    }
    /* Print the description of selected item */
    mcd = menuItems[current].desc;
    mvaddstr( WARNING_START + 1, (COLUMNS - strlen( mcd )) / 2, mcd );
    return y;
}

/* This function takes a list of menu items, lets the user choose one *
 * and returns the value keyboard shortcut of the selected menu item  */

int menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
		const char *available, int menuType, int menuDefault )
{
    int i, ylast = y, key = 0, current = menuDefault;
    if( !( menuType & ( MENU_HORIZ | MENU_VERT ) ) )    
    {
	doprintf("Menu without direction. Defaulting horizontal.");
	menuType |= MENU_HORIZ;
    }
    /* Make sure that the current is one of the available items */
    while( !strchr(available, menuItems[current].key) )
    {
	current ++ ;
	if( !menuItems[current].key ) current = 0;
    }
    /* Repeat until allowable choice has been made */
    while( !key )
    {
	/* Display the menu */
	ylast = menuUpdate( y, x, menuItems, itemLength, available,
			    menuType, current );
	refresh();
	key = getch();

	/* Clear out all prompts and such */
//        clear_warning();
	for( i = y; i < ylast; i ++ )
	{
	    move( i, x );
	    clrtoeol();
	}
	move( WARNING_START + 1, 0 );
	clrtoeol();
	/* Cursor keys */
#ifdef DJGPP
		switch(key)
#else
	if( key == ESC )
	{
	    /* Check whether this is a real ESC or one of extended keys */
	    /*nodelay(stdscr, TRUE);*/
	    key = getch();
	    /*nodelay(stdscr, FALSE);*/
	    if( key == /*ERR*/ ESC )
	    {
		/* This is a real ESC */
		key = ESC;
	    }
	    if( key == '[' )
	    {
		/* This is one extended keys */
		switch( getch() )
#endif
		{
		  case 'A': /* Up arrow */
		  case 0x103:
			if( menuType & MENU_VERT )
			{
			  do {
				current -- ;
				if( current < 0 ) while( menuItems[current+1].key ) current ++ ;
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_UP;
			break;
		  case 'B': /* Down arrow */
		  case 0x102:
			if( menuType & MENU_VERT )
			{
			  do {
				current ++ ;
				if( !menuItems[current].key ) current = 0 ;
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_DOWN;
			break;
		  case 'C': /* Right arrow */
		  case 0x105:
			if( menuType & MENU_HORIZ )
			{
			  do {
				current ++ ;
				if( !menuItems[current].key ) 
				{
				  current = 0 ;
				}
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_RIGHT;
			break;
		  case 'D': /* Left arrow */
		  case 0x104:
			if( menuType & MENU_HORIZ )
			{
			  do {
				current -- ;
				if( current < 0 )
				{
				  while( menuItems[current + 1].key ) current ++ ;
				}
			  } while( !strchr( available, menuItems[current].key ) );
			  key = 0;
			}
			else
			  key = MENU_LEFT;
			break;
		}
#ifndef DJGPP
	    }
	}
#endif
	/* Enter equals to the keyboard shortcut of current menu item */
	if((key==13) || (key==10))
	{
	    key = menuItems[current].key;
	}
	/* Should all keys to be accepted? */
	if( key && (menuType & MENU_ACCEPT_OTHERS) ) break;
	/* Is pressed key among acceptable ones */
	if( key && (strchr(available, tolower(key)) || strchr(available, key)))
	    break;
	/* The key has not been accepted so far -> let's reject it */
#ifdef DEBUG
	if( key )
	{
		move(5,0);
		doprintf("key %03X",key);
	    key = 0;
	    putchar( BELL );
//            print_warning("Illegal key");
	}
#endif
    }
    /* Clear out prompts and such */
//    clear_warning();
    for( i = y; i <= ylast; i ++ )
    {
	move( i, x );
	clrtoeol();
    }
    move( WARNING_START + 1, 0 );
    clrtoeol();
    return key;
}

/* A function which displays "Press a key to continue"                  *
 * and waits for a keypress.                                            *
 * Perhaps calling function menuSelect is a bit overkill but who cares? */

void menuContinue(void)
{
    static struct MenuItem menuContinueBtn[]=
    {
	{ 'c', "", "Press a key to continue" },
	{ 0, NULL, NULL }
    };

    menuSelect(23, 0,
	menuContinueBtn, 0, "c", MENU_HORIZ | MENU_ACCEPT_OTHERS, 0 );
}

/* Function menuSelect takes way too many parameters  *
 * Luckily, most of time we can do with this function */

int menuSimple(struct MenuItem *menuItems, int menuDefault)
{
    int i, j, itemLength = 0;
    char available[MENU_MAX_ITEMS];

    for(i = 0; menuItems[i].key; i++)
    {
	j = strlen(menuItems[i].name);
	if( j > itemLength ) itemLength = j;
	available[i] = menuItems[i].key;
    }
    available[i] = 0;
    return menuSelect(18, 0, menuItems, itemLength, available, MENU_HORIZ | MENU_BUTTON, menuDefault);
}

/* End of command menu support code */

void change_geometry(t_param_disk param_disk)
{
  int done = FALSE;
  char def[LINE_LENGTH];
  char response[LINE_LENGTH];
  int tmp_val;
  int command;

  while (!done) {
	static struct MenuItem menuGeometry[]=
	{
	  { 'c', "Cylinders", "Change cylinder geometry" },
	  { 'h', "Heads", "Change head geometry" },
	  { 's', "Sectors", "Change sector geometry" },
	  { 'd', "Done", "Done with changing geometry" },
	  { 0, NULL, NULL }
	};
	move(COMMAND_LINE_Y, COMMAND_LINE_X);
	clrtoeol();
	refresh();
	command=menuSimple(menuGeometry, 3);
	command=toupper(command);
	switch (command) {
	  case 'C':
		sprintf(def, "%d", param_disk->CHS.cylinder+1);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of cylinders: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_CYLINDERS) {
			param_disk->CHS.cylinder = tmp_val-1;
		  } else
			doprintf("Illegal cylinders value");
		}
		break;
	  case 'H':
		sprintf(def, "%d", param_disk->CHS.head+1);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of heads: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_HEADS) {
			param_disk->CHS.head = tmp_val-1;
		  } else
			doprintf("Illegal heads value");
		}
		break;
	  case 'S':
		sprintf(def, "%d", param_disk->CHS.sector);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of sectors per track: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= MAX_SECTORS) {
			param_disk->CHS.sector = tmp_val;
		  } else
			doprintf("Illegal sectors value");
		}
		break;
	  case ESC:
	  case 'D':
		done = TRUE;
		break;
	  default:
		putchar(BELL);
		break;
	}
  }
}

void aff_copy_full()
{
  clear();
  mvaddstr(0,0,m_Copyright);
#ifdef DJGPP
  doprintf("Dos version\n\n");
#elif defined(BSD)
  doprintf("BSD version\n\n");
#elif defined(LINUX)
  doprintf("Linux version\n\n");
#else
  doprintf("Undefined OS\n\n");
#endif
}

void aff_copy()
{
  clear();
  mvaddstr(0,0,m_Copyright);
}

void do_curses_testdisk(int debug,int paranoid,int dump_ind, int fast_mode,int align,int do_analyse)
{
  int i;
  int command;
  int nbr_disk;
  int current_disk=0;
  int done=0;
  t_param_disk param_disk[NBR_DISK_MAX];
//  t_diskext tab_part[NBR_PART_MAX];
  static struct MenuItem menuMain[]=
  {
	{'a',"Analyse","Analyse current partition structure and recover lost partitions"},
	{'d',"Delete","Delete all partitions"},
	{'c',"MBR Code","Write a classic MBR code"},
	{'g',"Geometry", "Change disk geometry (expert only)" },
	{'o',"Options","Modify options"},
	{'e',"Editor","Basic disk editor"},
	{'q',"Quit","Quit program"},
	{0,NULL,NULL}
  };
  initscr();
  noecho();
#ifdef DJGPP
  keypad(stdscr, TRUE); // Need it to get arrow key
#else
  nonl(); //don't use for Dos version but enter will work with it... dilema
#endif
  crmode();
  //  intrflush(stdscr, FALSE);
  cbreak();
  for(i=0;i<NBR_DISK_MAX;i++)
  {
	param_disk[i]=(t_param_disk)MALLOC(sizeof(*param_disk[i]));
#ifndef DJGPP
	param_disk[i]->name=NULL;
#endif
  }
  nbr_disk=hd_parse(param_disk,debug);
  /* save disk parameters to rapport */
  ecrit_rapport("Hard disk list\n");
  for(i=0;i<nbr_disk;i++)
    aff_disk_rapport(param_disk[i]);
  /* ncurses interface */
  if(nbr_disk>0)
  {
    while(!done)
    {
      aff_copy_full();
      move(5,0);
      for(i=0;i<nbr_disk;i++)
      {
	if(i!=current_disk)
	  aff_disk(param_disk[i]);
	else
	{
	  standout();
	  aff_disk(param_disk[i]);
	  standend();
	}
      }
      refresh();
      command = menuSelect(INTER_MAIN_Y, INTER_MAIN_X, menuMain, 8,
	  "agdcoq", MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, 0);
      /* e for editor will be added when the editor will be better */
      switch(command)
      {
	case 'g':
	case 'G':
	  change_geometry(param_disk[current_disk]);
	  break;
	case 'd':
	case 'D':
	  move(INTER_MAIN_Y,INTER_MAIN_X);
	  write_clean_table(param_disk[current_disk]);
	  break;
	case 'c':
	case 'C':
	  move(INTER_MAIN_Y,INTER_MAIN_X);
	  write_MBR_code(param_disk[current_disk]);
	  break;
	case 'a':
	case 'A':
	  if(do_analyse)
	    interface_analyse(param_disk[current_disk],debug);
	  interface_recovery(param_disk[current_disk],paranoid,debug, dump_ind, fast_mode,align);
	  break;
	case 'e':
	case 'E':
	  interface_editor2(param_disk[current_disk]);
	  break;
	case 'q':
	case 'Q':
	  done = 1;
	  break;
	case 'o':
	case 'O':
	  interface_options(&paranoid,&dump_ind,&fast_mode,&align);
	  break;
	case MENU_UP:
	  if(current_disk>0)
	    current_disk--;
	  break;
	case MENU_DOWN:
	  if(current_disk<nbr_disk-1)
	    current_disk++;
	  break;
#ifdef DEBUG
	default:
	  move(10,0);
	  doprintf("key %03X %d", command, command);
	  refresh();
	  break;
#endif
      }
    }
  }
  endwin();
  if(debug)
	ecrit_rapport("main: free the memory\n");
  for(i=0;i<NBR_DISK_MAX;i++)
  {
#ifndef DJGPP
	if(param_disk[i]->name)
	  FREE(param_disk[i]->name);
#endif
	FREE(param_disk[i]);    /* forget name */
  }
}

void interface_analyse(t_param_disk_cst param_disk, const int debug)
{
  /* log */
  ecrit_rapport("\nAnalyse ");
  aff_disk_rapport(param_disk);
  /* ncurses interface */
  aff_copy();
  mvaddstr(23,0,m_PART_TYPE);
  move(4,0);
  aff_disk(param_disk);
  mvaddstr(5,0,"Check current partition structure");
  move(6,0);
  refresh();
  test_MBR(param_disk,debug);
}

void interface_recovery(t_param_disk_cst param_disk,const int paranoid, const int debug, const int dump_ind, const int fast_mode, const int align)
{
  aff_copy();
  move(4,0);
  aff_disk(param_disk);
  move(5,0);
  god_mode(param_disk,paranoid,debug,dump_ind,fast_mode,align);
}

int interface_write(t_param_disk_cst param_disk,t_list_part list_part)
{
  t_list_part parts;
  ecrit_rapport("\ninterface_write()\n");
  aff_copy();
  move(4,0);
  aff_disk(param_disk);
  move(5,0);
  aff_buffer(BUFFER_RESET,"");
  mvaddstr(6,0,m_PART_HEADER_WRITE);
  for(parts=list_part;parts;parts=parts->next)
	if(parts->part->status!=STATUS_LOG)
	  aff_part_buffer(2,param_disk,parts->part);
  for(parts=list_part;parts;parts=parts->next)
	if(parts->part->status==STATUS_LOG)
	  aff_part_buffer(2,param_disk,parts->part);
  return aff_buffer(BUFFER_DISPLAY,"W");
}

void interface_options(int *paranoid,int *dump_ind,int *fast_mode,int *align)
{
/* paranoid: search in partition
 * dump_ind dump sector Y/N
 * fast_mode Partitions are aligned on cylinder boundary Y/N
 * */
  int done = FALSE;
  int menu = 3;
  // write old options to log file
  ecrit_rapport("Old options :\n%s\n%s\n%s\n",
	*paranoid?"Paranoid : Yes":"Paranoid : No",
	*dump_ind?"Dump : Yes":"Dump : No",
	*align?"Cylinder boundary : Yes":"Cylinder boundary : No",
	*fast_mode?"Fast mode : Yes":"Fast mode : No");
  // ncurses interface
  while (!done) {
	struct MenuItem menuOptions[]=
	{
	  { 'P', NULL, "Search partition" },
	  { 'D',NULL,"Dump essential sectors" },
	  { 'C',NULL,"Partitions are aligned on cylinder boundary" },
	  { 'F',NULL,"Don't check every hard disk head" },
	  { 'Q',"Quit","Return to main menu"},
	  { 0, NULL, NULL }
	};
	menuOptions[0].name=*paranoid?"Paranoid : Yes":"Paranoid : No";
	menuOptions[1].name=*dump_ind?"Dump : Yes":"Dump : No";
	menuOptions[2].name=*align?"Cylinder boundary : Yes":"Cylinder boundary : No";
	menuOptions[3].name=*fast_mode?"Fast mode : Yes":"Fast mode : No";
//      move(COMMAND_LINE_Y, COMMAND_LINE_X);
//      clrtoeol();
	refresh();
    switch (toupper( menuSelect(INTER_OPTION_Y, INTER_OPTION_X, menuOptions, 0, "PDCFQ", MENU_VERT, menu)))
	{
	  case 'P':
		*paranoid=!*paranoid;
		menu=0;
		break;
	  case 'D':
		*dump_ind=!*dump_ind;
		menu=1;
		break;
	  case 'C':
		*align=!*align;
		menu=2;
		break;
	  case 'F':
		*fast_mode=!*fast_mode;
		menu=3;
		break;
	  case ESC:
	  case 'Q':
		done = TRUE;
		break;
	}
  }
  // write new options to log file
  ecrit_rapport("New options :\n%s\n%s\n%s\n",
	*paranoid?"Paranoid : Yes":"Paranoid : No",
	*dump_ind?"Dump : Yes":"Dump : No",
	*align?"Cylinder boundary : Yes":"Cylinder boundary : No",
	*fast_mode?"Fast mode : Yes":"Fast mode : No");
}

void dump(void *nom_dump,uint lng)
{
  unsigned int i,j;
  unsigned int pos=0;
  uchar car;
  int done=0;
  int menu=2;   /* default : quit */
  struct MenuItem menuDump[]=
  {
	{ 'P', "Previous",""},
	{ 'N', "Next","" },
	{ 'Q',"Quit","Quit dump section"},
	{ 0, NULL, NULL }
  };
  /* write dump to log file*/
  for (i=0; (i<lng/0x10); i++)
  {
	ecrit_rapport("%04X ",i*0x10);
	for(j=0; j< 0x10;j++)
	{
	  car=*((uchar *)nom_dump+i*0x10+j);
	  ecrit_rapport("%02x", car);
	  if(j%4==(4-1))
		ecrit_rapport(" ");
	}
	ecrit_rapport("  ");
	for(j=0; j< 0x10;j++)
	{
	  car=*((uchar*)nom_dump+i*0x10+j);
	  if ((car<32)||(car >= 127))
		ecrit_rapport(" ");
	  else
		ecrit_rapport("%c",  car);
	}
	ecrit_rapport("\n");
  }
  /* ncurses interface */
  mvaddstr(DUMP_Y,DUMP_X,m_DUMP_HEXA);
  /* On pourrait utiliser wscrl */
  do
  {
	for (i=pos; (i<lng/0x10)&&((i-pos)<DUMP_MAX_LINES); i++)
	{
	  move(DUMP_Y+1+i-pos,DUMP_X);
	  doprintf("%04X ",i*0x10);
	  for(j=0; j< 0x10;j++)
	  {
		car=*((uchar*)nom_dump+i*0x10+j);
		doprintf("%02x", car);
		if(j%4==(4-1))
		  doprintf(" ");
	  }
	  doprintf("  ");
	  for(j=0; j< 0x10;j++)
	  {
		car=*((uchar*)nom_dump+i*0x10+j);
		if ((car<32)||(car >= 127))
		  doprintf(" ");
		else
		  doprintf("%c",  car);
	  }
	}
	switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuDump, 8, "PNQ", MENU_HORIZ | MENU_ACCEPT_OTHERS, menu)))
	{
	  case 'P':
	  case MENU_UP:
		menu=0;
		if(pos>0)
		  pos--;
		break;
	  case 'N':
	  case MENU_DOWN:
		menu=1;
		if(pos<lng/0x10-DUMP_MAX_LINES+1)
		  pos++;
		break;
	  case ESC:
	  case 'Q':
		done = TRUE;
		break;
	}
  } while(!done);
}

int dump2(const uchar *nom_dump,uint lng)
{
  unsigned int i,j;
  unsigned int pos=0;
  uchar car;
  int done=0;
  int menu=2;   /* default : quit */
  struct MenuItem menuDump[]=
  {
	{ 'P', "Previous",""},
	{ 'N', "Next","" },
	{ 'Q',"Quit","Quit dump section"},
	{ 0, NULL, NULL }
  };
  /* write dump to log file*/
  for (i=0; (i<lng/0x10); i++)
  {
	ecrit_rapport("%04X ",i*0x10);
	for(j=0; j< 0x10;j++)
	{
	  car=*(nom_dump+i*0x10+j);
	  ecrit_rapport("%02x", car);
	  if(j%4==(4-1))
		ecrit_rapport(" ");
	}
	ecrit_rapport("  ");
	for(j=0; j< 0x10;j++)
	{
	  car=*(nom_dump+i*0x10+j);
	  if ((car<32)||(car >= 127))
		ecrit_rapport(" ");
	  else
		ecrit_rapport("%c",  car);
	}
	ecrit_rapport("\n");
  }
  /* ncurses interface */
  mvaddstr(DUMP_Y,DUMP_X,m_DUMP_HEXA);
  do
  {
	for (i=pos; (i<lng/0x10)&&((i-pos)<DUMP_MAX_LINES); i++)
	{
	  move(DUMP_Y+1+i-pos,DUMP_X);
	  doprintf("%04X ",i*0x10);
	  for(j=0; j< 0x10;j++)
	  {
		car=*(nom_dump+i*0x10+j);
		doprintf("%02x", car);
		if(j%4==(4-1))
		  doprintf(" ");
	  }
	  doprintf("  ");
	  for(j=0; j< 0x10;j++)
	  {
		car=*(nom_dump+i*0x10+j);
		if ((car<32)||(car >= 127))
		  doprintf(" ");
		else
		  doprintf("%c",  car);
	  }
	}
	switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuDump, 8, "PNQ", MENU_HORIZ | MENU_ACCEPT_OTHERS, menu)))
	{
	  case 'P':
	  case MENU_UP:
		menu=0;
		if(pos>0)
		  pos--;
		else
		  done=MENU_UP;
		break;
	  case 'N':
	  case MENU_DOWN:
		menu=1;
		if(pos<lng/0x10-DUMP_MAX_LINES+1)
		  pos++;
		else
		  done = MENU_DOWN;
		break;
	  case ESC:
	  case 'Q':
		done = 'Q';
		break;
	}
  } while(!done);
  return done;
}

int aff_buffer(const int cmd, const char *_format, ...)
{
  static char buffer[MAX_LINES][LINE_LENGTH+1];
  static int nbr_lines=0;
  switch(cmd)
  {
	case BUFFER_RESET:
	  {
		int i;
		nbr_lines=0;
		for(i=0;i<MAX_LINES;i++)
		  memset(buffer[i],0,LINE_LENGTH+1);
	  }
	  break;
	case BUFFER_ADD:
	  {
		char tmp_line[BUFFER_LINE_LENGTH+1];
		char *pos_in_tmp_line=tmp_line-1;
		memset(tmp_line,'\0',sizeof(tmp_line));
#ifdef DJGPP
		if(strlen(_format)>sizeof(tmp_line))
		  ecrit_rapport("\nBUG: %s\n",_format);
		vsprintf(tmp_line,_format,(&_format)+1);
#else
		vsnprintf(tmp_line,BUFFER_LINE_LENGTH,_format,(va_list)(&_format+1));
#endif
		refresh();

		while(pos_in_tmp_line && (nbr_lines<MAX_LINES))
		{
		  char *ret_ligne= strchr(++pos_in_tmp_line,'\n');
		  if(ret_ligne)
		  {
			strncat(buffer[nbr_lines],pos_in_tmp_line,
				ret_ligne-pos_in_tmp_line<LINE_LENGTH-strlen(buffer[nbr_lines])?ret_ligne-pos_in_tmp_line:LINE_LENGTH-strlen(buffer[nbr_lines]));
			if(strlen(buffer[nbr_lines])>0)
			  nbr_lines++;
		  }
		  else
		  {
			strncat(buffer[nbr_lines],pos_in_tmp_line,LINE_LENGTH-strlen(buffer[nbr_lines]));
		  }
		  pos_in_tmp_line=ret_ligne;
		}
//		ecrit_rapport("aff_buffer %d =>%s<=\n",nbr_lines,tmp_line);
		if(nbr_lines>=MAX_LINES)
		{
		  ecrit_rapport("aff_buffer too much lines =>%s<=\n",tmp_line);
		}
	  }
	  break;
	case BUFFER_DISPLAY:
	  {
		int i;
		int menu;
		int pos=0;
		int done=0;
		struct MenuItem menuBuffer[]=
		{
		  { 'P', "Previous",""},
		  { 'N', "Next","" },
		  { 'Q',"Quit","Quit this section"},
		  { 'W', "Write","Write partition structure to disk"},
		  { 0, NULL, NULL }
		};
		char options[20];
		if(nbr_lines<=DUMP_MAX_LINES)
		{
		  strcpy(options,"Q");
		  menu=0;       /* default = quit */
		}
		else
		{
		  strcpy(options,"PNQ");
		  menu=1;       /* default = next */
		}
		strncat(options,_format,sizeof(options)-strlen(options));
		if(buffer[nbr_lines][0])
		  nbr_lines++;
		/* to log file */
		for(i=0;i<nbr_lines;i++)
		  ecrit_rapport("%s\n",buffer[i]);
		/* curses interface */
		do
		{
		  for (i=pos; (i<nbr_lines)&&((i-pos)<DUMP_MAX_LINES); i++)
		  {
			move(DUMP_Y+1+i-pos,DUMP_X);
			clrtoeol();
			doprintf("%s",buffer[i]);
		  }
		  
		  switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuBuffer, 8, options, MENU_HORIZ | MENU_ACCEPT_OTHERS, menu)))
		  {
			case 'P':
			case MENU_UP:
			  menu=0;
			  if(pos>0)
				pos--;
			  break;
			case 'N':
			case MENU_DOWN:
			  menu=1;
			  if(pos<nbr_lines-DUMP_MAX_LINES+1)
				pos++;
			  break;
			case 'W':
			  if(strchr(options,'W'))
				return 1;
			case ESC:
			case 'Q':
			  done = TRUE;
			  break;
		  }
		} while(!done);
	  }
	  break;
  }
  return 0;
}

void aff_part_buffer(const int newline,t_param_disk_cst disk_car,t_diskext_cst partition)
{
  switch(newline)
  {
    case 1:
      aff_buffer(BUFFER_ADD,"\n   ");
      break;
    case 2:
      if(partition->status!=STATUS_EXT_IN_EXT)
	aff_buffer(BUFFER_ADD,"%2d ", partition->order);
      else
	aff_buffer(BUFFER_ADD,"   ");
      break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           aff_buffer(BUFFER_ADD,"P"); break;
    case STATUS_PRIM_BOOT:      aff_buffer(BUFFER_ADD,"*"); break;
    case STATUS_EXT:            aff_buffer(BUFFER_ADD,"E"); break;
    case STATUS_EXT_IN_EXT:     aff_buffer(BUFFER_ADD,"X"); break;
    case STATUS_LOG:            aff_buffer(BUFFER_ADD,"L"); break;
    case STATUS_DELETED:        aff_buffer(BUFFER_ADD,"D"); break;
  }
  if(nom_os[partition->part_type])
    aff_buffer(BUFFER_ADD," %-28s ",nom_os[partition->part_type]);
  else
    aff_buffer(BUFFER_ADD," Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_from_LBA_buffer(disk_car,partition->lba);
  aff_CHS_from_LBA_buffer(disk_car,partition->lba+partition->part_size-1);
  if(partition->sect_rel!=0)
    aff_buffer(BUFFER_ADD,"%10lu ", partition->sect_rel);
  aff_buffer(BUFFER_ADD,"%10lu", partition->part_size);
  if(newline==2)
    aff_buffer(BUFFER_ADD,"\n");
}

void aff_entree(t_entree_cst entree)
{
  if(nom_os[os(entree)])
    doprintf("\n %-28s ",nom_os[os(entree)]);
  else
    doprintf("\n Sys=%02X\t\t\t  ", os(entree));
  doprintf("%4u %3u %2u"
	 " %4u %3u %2u"
	 " %10lu"
	 " %10lu",
	  s_cyl(entree),s_head(entree),s_sect(entree),
	  e_cyl(entree),e_head(entree),e_sect(entree),
	  sect_rel(entree),nbr_sect(entree));
}

void aff_entree_rapport(t_entree_cst entree)
{
  if(nom_os[os(entree)])
    ecrit_rapport(" %-28s ",nom_os[os(entree)]);
  else
    ecrit_rapport(" Sys=%02X\t\t\t  ", os(entree));
  ecrit_rapport("%4u %3u %2u"
	 " %4u %3u %2u"
	 " %10lu"
	 " %10lu\n",
	  s_cyl(entree),s_head(entree),s_sect(entree),
	  e_cyl(entree),e_head(entree),e_sect(entree),
	  sect_rel(entree),nbr_sect(entree));
}

void aff_CHS_rapport(const t_CHS * CHS)
{
  ecrit_rapport("%4u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS(const t_CHS * CHS)
{
  doprintf("%4u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS_buffer(const t_CHS * CHS)
{
  aff_buffer(BUFFER_ADD,"%4u %3u %2u ", CHS->cylinder, CHS->head, CHS->sector);
}

void aff_CHS_from_LBA(t_param_disk_cst disk_car, dword hd_offset)
{
  doprintf("%4u %3u %2u ", LBA2cylinder(disk_car,hd_offset), LBA2head(disk_car,hd_offset),LBA2sector(disk_car,hd_offset));
}

void aff_CHS_from_LBA_buffer(t_param_disk_cst disk_car, dword hd_offset)
{
  aff_buffer(BUFFER_ADD,"%4u %3u %2u ", LBA2cylinder(disk_car,hd_offset), LBA2head(disk_car,hd_offset),LBA2sector(disk_car,hd_offset));
}

void aff_CHS_from_LBA_rapport(t_param_disk_cst disk_car, dword hd_offset)
{
  ecrit_rapport("%4u %3u %2u ", LBA2cylinder(disk_car,hd_offset), LBA2head(disk_car,hd_offset),LBA2sector(disk_car,hd_offset));
}

void aff_part_rapport(const int newline,t_param_disk_cst disk_car,t_diskext_cst partition)
{
  switch(newline)
  {
    case 1:
      ecrit_rapport("\n   ");
      break;
    case 2: 
      if(partition->status!=STATUS_EXT_IN_EXT)
	ecrit_rapport("%2d ", partition->order);
      else
	ecrit_rapport("   ");
      break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           ecrit_rapport("P"); break;
    case STATUS_PRIM_BOOT:      ecrit_rapport("*"); break;
    case STATUS_EXT:            ecrit_rapport("E"); break;
    case STATUS_EXT_IN_EXT:     ecrit_rapport("X"); break;
    case STATUS_LOG:            ecrit_rapport("L"); break;
    case STATUS_DELETED:        ecrit_rapport("D"); break;
  }
  if(nom_os[partition->part_type])
    ecrit_rapport(" %-28s ",nom_os[partition->part_type]);
  else
    ecrit_rapport(" Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_from_LBA_rapport(disk_car,partition->lba);
  aff_CHS_from_LBA_rapport(disk_car,partition->lba+partition->part_size-1);
  if(partition->sect_rel!=0)
    ecrit_rapport("%10lu ", partition->sect_rel);
  ecrit_rapport("%10lu", partition->part_size);
  if(newline==2)
    ecrit_rapport("\n");
}

void aff_part(const int newline,t_param_disk_cst disk_car,t_diskext_cst partition)
{
  switch(newline)
  {
    case 1:
      doprintf("\n   ");
      break;
    case 2:
      if(partition->status!=STATUS_EXT_IN_EXT)
	doprintf("%2d ", partition->order);
      else
	doprintf("   ");
      break;
  }
  switch(partition->status)
  {
    case STATUS_PRIM:           doprintf("P"); break;
    case STATUS_PRIM_BOOT:      doprintf("*"); break;
    case STATUS_EXT:            doprintf("E"); break;
    case STATUS_EXT_IN_EXT:     doprintf("X"); break;
    case STATUS_LOG:            doprintf("L"); break;
    case STATUS_DELETED:        doprintf("D"); break;
  }
  if(nom_os[partition->part_type])
    doprintf(" %-28s ",nom_os[partition->part_type]);
  else
    doprintf(" Sys=%02X\t\t\t  ", partition->part_type);
  aff_CHS_from_LBA(disk_car,partition->lba);
  aff_CHS_from_LBA(disk_car,partition->lba+partition->part_size-1);
  if(partition->sect_rel!=0)
    doprintf("%10lu ", partition->sect_rel);
  doprintf("%10lu", partition->part_size);
  refresh();
}

void aff_LBA2CHS_rapport(t_param_disk_cst disk_car,dword pos_LBA)
{
  dword tmp;
  dword cylinder, head, sector;
  tmp=disk_car->CHS.sector;
  sector=(pos_LBA%tmp)+1;
  tmp=pos_LBA/tmp;
  cylinder=tmp/(disk_car->CHS.head+1);
  head=tmp%(disk_car->CHS.head+1);
  ecrit_rapport("%d/%d/%d", cylinder,head,sector);
}

void aff_disk(t_param_disk_cst param_disk)
{
#ifdef DJGPP
  doprintf(m_DISK " %02x - CHS %d %d %d - %d MB (%s BIOS mode)\n",
      param_disk->disk, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024,param_disk->mode==0?"Std":param_disk->mode==1?"Enh":"Mxd");
#else
  doprintf(m_DISK " %s - CHS %d %d %d - %d MB\n",
      param_disk->name, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024);
#endif
}

void aff_disk_rapport(t_param_disk_cst param_disk)
{
#ifdef DJGPP
  ecrit_rapport(m_DISK " %02x - CHS %d %d %d - %d MB (%s BIOS mode)\n",
      param_disk->disk, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024,param_disk->mode?"Enh":"Std");
#else
  ecrit_rapport(m_DISK " %s - CHS %d %d %d - %d MB\n",
      param_disk->name, param_disk->CHS.cylinder+1, param_disk->CHS.head+1, param_disk->CHS.sector,(param_disk->CHS.cylinder+1)*(param_disk->CHS.head+1)*param_disk->CHS.sector/2/1024);
#endif
}

int ask_structure(t_param_disk_cst disk_car,t_list_part list_part)
{
  int quit;
  int offset=0;
  int pos_num=0;
  t_list_part pos=list_part;
  aff_copy();
  move(4,0);
  aff_disk(disk_car);
  mvaddstr(23,0,m_PART_TYPE);
  mvaddstr(5,0,m_CHOOSE_PART_TYPE);
  mvaddstr(6,0,m_PART_HEADER2);
  do
  {
    int i;
    t_list_part parts;
    int car;
    for(i=0,parts=list_part;parts && (i<offset);parts=parts->next,i++);
    for(i=offset;parts &&((i-offset)<INTER_STRUCTURE);i++,parts=parts->next)
    {
      move(5+2+i-offset,0);
      if(parts==pos)
	standout();
      aff_part(0,disk_car,parts->part);
      if(parts==pos)
	standend();
    }
    if(test_structure(list_part))
      mvaddstr(22,0,m_STRUCT_BAD);
    else
      mvaddstr(22,0,m_STRUCT_OK);
    refresh();
    car=getch();
    quit=0;
    switch(car)
    {
      case 0x103:
      case 'A':
	only_one_bootable(list_part,pos);
	if(pos->prev)
	{
	  pos=pos->prev;
	  pos_num--;
	}
	if(pos_num<offset)
	  offset--;
	break;
      case 0x102:
      case 'B':
	only_one_bootable(list_part,pos);
	if(pos->next)
	{
	  pos=pos->next;
	  pos_num++;
	}
	if(pos_num>=offset+INTER_STRUCTURE)
	  offset++;
	break;
      case 0x105:	 /* right */
      case 'C':
      case '+':
      case ' ':
      case 'l':
	if(pos->part->status==STATUS_LOG)
	  pos->part->status=STATUS_DELETED;
	else
	  pos->part->status++;
	if(pos->part->status==STATUS_LOG && !can_be_ext(disk_car,pos->part))
	  pos->part->status=STATUS_DELETED;
	break;
      case 0x104:	/* left */
      case 'D':
      case '-':
      case 'S':
	if(pos->part->status==STATUS_DELETED)
	  pos->part->status=STATUS_LOG;
	else
	  pos->part->status--;
	if(pos->part->status==STATUS_LOG && !can_be_ext(disk_car,pos->part))
	  pos->part->status--;
	break;
      case 'q':
      case '\r':
      case '\n':
      case 'M':
	quit=1;
	break;
      default:
//      doprintf("Touche %02X", car);
	break;
    }
  } while(!quit ||test_structure(list_part));
  return 0;
}

void aff_MBR_data(t_param_disk_cst disk_car,const uint nbr_part, const t_diskext tab_part[NBR_PART_MAX])
{
  uint i;
  doprintf(m_PART_HEADER);
//  for(parts=list_part;parts;parts=parts->next)
  for(i=0;i<nbr_part;i++)
  {
    aff_part_buffer(2,disk_car,tab_part[i]);
    if(tab_part[i]->errcode)
      aff_buffer(BUFFER_ADD,"%s\n",errmsg_entree2partition(tab_part[i]->errcode));
  }
}

void interface_editor(t_param_disk_cst param_disk)
{
  int done = FALSE;
  loff_t hd_offset=0;
  aff_disk_rapport(param_disk);
  while (!done)
  {
    static struct MenuItem menuEditor[]=
    {
      { 'c', "Change position", "" },
      { 'd', "Dump", "Dump sector" },
      { 'q', "Quit",""},
      { 0, NULL, NULL }
    };
    move(COMMAND_LINE_Y, COMMAND_LINE_X);
    clrtoeol();
    refresh();
    switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuEditor, 8, "cdq", MENU_HORIZ | MENU_ACCEPT_OTHERS, 0)))
    {
      case 'C':
//	interface_editor_position(param_disk,&hd_offset);
	break;
      case 'D':
	{
	  uchar buffer[0x200];
	  if(hd_read_LBA(param_disk,1, buffer, hd_offset))
	    doprintf(m_PART_RD_ERR);
	  else
	    dump(buffer,0x200);
	}
	break;
      case ESC:
      case 'Q':
	done = TRUE;
	break;
    }
  }
}

void interface_editor2(t_param_disk_cst param_disk)
{
  int done = FALSE;
  loff_t hd_offset=0;
  aff_disk_rapport(param_disk);
  while (!done)
  {
    static struct MenuItem menuEditor[]=
    {
      { 'c', "Change position", "" },
      { 'd', "Dump", "Dump sector" },
      { 'q', "Quit",""},
      { 0, NULL, NULL }
    };
    move(COMMAND_LINE_Y, COMMAND_LINE_X);
    clrtoeol();
    refresh();
    switch (toupper( menuSelect(INTER_DUMP_Y, INTER_DUMP_X, menuEditor, 8, "cdq", MENU_HORIZ | MENU_ACCEPT_OTHERS, 0)))
    {
      case 'C':
//	interface_editor_position(param_disk,&hd_offset);
	break;
      case 'D':
	{
	  uchar buffer[0x200];
	  while(!done)
	  {
	    move(4,0);
	    clrtoeol();
	    doprintf("%ld", hd_offset);
	    if(hd_read_LBA(param_disk,1, buffer, hd_offset))
	      doprintf(m_PART_RD_ERR);
	    else
	    {
	      switch(dump2(buffer,0x200))
	      {
		case MENU_UP:
		  if(hd_offset>0)
		    hd_offset--;
		  break;
		case MENU_DOWN:
		  if(hd_offset<param_disk->lba-1)
		    hd_offset++;
		  break;
		default:
		  done = TRUE;
		  break;
	      }
	    }
	  }
	}
	break;
      case ESC:
      case 'Q':
	done = TRUE;
	break;
    }
  }
}

void interface_editor_position(t_param_disk_cst param_disk, t_disk position)
{
  int done = FALSE;
  char def[LINE_LENGTH];
  char response[LINE_LENGTH];
  int tmp_val;
  int command;

  while (!done) {
	static struct MenuItem menuGeometry[]=
	{
	  { 'c', "Cylinders", "Change cylinder" },
	  { 'h', "Heads", "Change head" },
	  { 's', "Sectors", "Change sector" },
	  { 'd', "Done", "Done with changing" },
	  { 0, NULL, NULL }
	};
	move(COMMAND_LINE_Y, COMMAND_LINE_X);
	clrtoeol();
	refresh();
	command=menuSimple(menuGeometry, 3);
	command=toupper(command);
	switch (command) {
	  case 'C':
		sprintf(def, "%d", position->cylinder);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of cylinders: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val >= 0 && tmp_val <= param_disk->CHS.cylinder) {
			position->cylinder = tmp_val;
		  } else
			doprintf("Illegal cylinders value");
		}
		break;
	  case 'H':
		sprintf(def, "%d", position->head);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of heads: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val >= 0 && tmp_val <= param_disk->CHS.head) {
			position->head = tmp_val;
		  } else
			doprintf("Illegal heads value");
		}
		break;
	  case 'S':
		sprintf(def, "%d", position->sector);
		mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X, "Enter the number of sectors per track: ");
		if (get_string(response, LINE_LENGTH, def) > 0) {
		  tmp_val = atoi(response);
		  if (tmp_val > 0 && tmp_val <= param_disk->CHS.sector ) {
			position->sector = tmp_val;
		  } else
			doprintf("Illegal sectors value");
		}
		break;
	  case ESC:
	  case 'D':
		done = TRUE;
		break;
	  default:
		putchar(BELL);
		break;
	}
  }
}
