/*

    File: interface.h

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
/* Constants for menuType parameter of menuSelect function */
#define MENU_HORIZ 1
#define MENU_VERT 2
#define MENU_ACCEPT_OTHERS 4
#define MENU_BUTTON 8
/* Miscellenous constants */
#define MENU_SPACING 2
#define MENU_MAX_ITEMS 256 /* for simpleMenu function */
#define MENU_UP 1
#define MENU_DOWN 2
#define MENU_RIGHT 3
#define MENU_LEFT 4

struct MenuItem
{
    const int key; /* Keyboard shortcut; if zero, then there is no more items in the menu item table */
    const char *name; /* Item name, should be eight characters with current implementation */
    const char *desc; /* Item description to be printed when item is selected */
};

#define LINE_LENGTH 80
#define BUFFER_LINE_LENGTH 4*LINE_LENGTH
#define MAXIMUM_PARTS 60
#define COL_ID_WIDTH 25
#define WARNING_START 23
#define COLUMNS 80

#define CR '\015'
#define ESC '\033'
#define DEL '\177'
#define BELL '\007'
/* '\014' == ^L */
#define REDRAWKEY '\014'


#define INTER_MAIN_X	0
#define INTER_MAIN_Y	18
#define COMMAND_LINE_X	0
#define COMMAND_LINE_Y  18
#define INTER_OPTION_X  0
#define INTER_OPTION_Y	COMMAND_LINE_Y
#define ANALYSE_X	0
#define ANALYSE_Y	5
#define DUMP_X	0
#define DUMP_Y	ANALYSE_Y + 1
#define DUMP_MAX_LINES	15
#define INTER_DUMP_X	DUMP_X
#define INTER_DUMP_Y	DUMP_Y+DUMP_MAX_LINES
#define INTER_STRUCTURE	15

int menuUpdate( int y, int x, const struct MenuItem *menuItems, const int itemLength, const char *available, const int menuType, int current );
int menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength, const char *available, int menuType, int menuDefault );
void menuContinue(void);
int menuSimple(struct MenuItem *menuItems, int menuDefault);
void change_geometry(t_param_disk param_disk);
void do_curses_testdisk(int debug,int paranoid, int dump_ind, int fast_mode, int align,int do_analyse);
int interface_write(t_param_disk_cst param_disk,t_list_part list_part);
//void dump(const uchar *nom_dump,uint lng);
void dump(void *nom_dump,uint lng);
int getch_nodelay(void);

enum {BUFFER_RESET, BUFFER_ADD, BUFFER_DISPLAY};
#define MAX_LINES	60
void aff_part_buffer(const int newline,t_param_disk_cst disk_car,t_diskext_cst partition);
int aff_buffer(const int cmd, const char *_format, ...);
void aff_LBA2CHS_rapport(t_param_disk_cst disk_car,dword pos_LBA);
void aff_disk(t_param_disk_cst param_disk);
void aff_disk_rapport(t_param_disk_cst param_disk);
void aff_CHS_rapport(const t_CHS * CHS);
void aff_CHS(const t_CHS * CHS);
void aff_CHS_buffer(const t_CHS * CHS);
void aff_entree(t_entree_cst);
void aff_entree_rapport(t_entree_cst);
void aff_part(const int newline,t_param_disk_cst disk_car,t_diskext_cst partition);
void aff_part_rapport(const int newline,t_param_disk_cst disk_car,t_diskext_cst partition);
int ask_structure(t_param_disk_cst disk_car,t_list_part list_part);
void aff_MBR_data(t_param_disk_cst disk_car,const uint nbr_part, const t_diskext tab_part[NBR_PART_MAX]);
void aff_copy(void);
void aff_copy_full(void);
